import discord
from discord.enums import ChannelType
from discord.errors import InvalidArgument, InvalidData
from .types import OptionType
from ..tools import get


class ParseMethod:
    """Methods of how the interaction argument data should be treated

    - ``RAW``       [0]
    : Returns the raw value which was received
    - ``RESOLVE``   [1]
    : Uses the resolved data which will be delivered together with the received interaction
    - ``FETCH``     [2]
    : Fetches all the ids of the received data with an api call
    - ``CACHE``     [3]
    : Uses the internal bot cache to receive the data ``.get_channel``, ``.get_guild``...

    .. warning::

        Some cache functions need to have intents enabled (more information https://discordpy.readthedocs.io/en/latest/intents.html?highlight=intents)

    - ``AUTO``      [4]
    : This will try to resolve the data, and if an exception occurs, it will try to fetch the data,
    and if that doesn't work it will try to get the data from the internal cache and if even this doesn't work, it will use raw data
    """
    RAW         =       0
    RESOLVE     =       1
    FETCH       =       2
    CACHE       =       3
    AUTO        =       4

def resolve(data, _discord):
    resolved = {}
    for x in data["data"]["resolved"]:
        if x == "members":
            resolved["members"] = {}
            for m_id in data["data"]["resolved"]["members"]:
                member_data = data["data"]["resolved"]["members"][m_id]
                member_data["user"] = data["data"]["resolved"]["users"][m_id]
                resolved["members"][m_id] = discord.Member(data=member_data, guild=_discord.get_guild(data["guild_id"]), state=_discord._connection)
        elif x == "messages":
            resolved["messages"] = {}
            for message_id in data["data"]["resolved"]["messages"]:
                message_data = data["data"]["resolved"]["messages"][message_id]
                resolved["messages"][message_id] = discord.Message(data=message_data, channel=_discord.get_channel(data["channel_id"]), state=_discord._connection)
        elif x == "channels":
            resolved["channels"] = {}
            for channel_id in data["data"]["resolved"]["channels"]:
                channel_data = data["data"]["resolved"]["channels"][channel_id]

                guild = _discord.get_guild(data["guild_id"])
                channel = None
                if ChannelType(channel_data["type"]) is  ChannelType.text:
                    channel = discord.TextChannel(data=channel_data, guild=guild, state=_discord._connection)
                elif ChannelType(channel_data["type"]) is  ChannelType.voice:
                    channel = discord.VoiceChannel(data=channel_data, guild=guild, state=_discord._connection)
                elif ChannelType(channel_data["type"]) is  ChannelType.category:
                    channel = discord.CategoryChannel(data=channel_data, guild=guild, state=_discord._connection)
                elif ChannelType(channel_data["type"]) is  ChannelType.group:
                    channel = discord.GroupChannel(data=channel_data, guild=guild, state=_discord._connection)
                elif ChannelType(channel_data["type"]) is  ChannelType.news:
                    channel = discord.NewsChannel(data=channel_data, guild=guild, state=_discord._connection)
                elif ChannelType(channel_data["type"]) is  ChannelType.private:
                    channel = discord.DMChannel(data=channel_data, guild=guild, state=_discord._connection)
                elif ChannelType(channel_data["type"]) is  ChannelType.store:
                    channel = discord.StoreChannel(data=channel_data, guild=guild, state=_discord._connection)
                elif ChannelType(channel_data["type"]) is  ChannelType.stage_voice:
                    channel = discord.StageChannel(data=channel_data, guild=guild, state=_discord._connection)
                resolved["channels"][channel_id] = channel
        elif x == "roles":
            resolved["roles"] = {}
            for role_id in data["data"]["resolved"]["roles"]:
                role_data = data["data"]["resolved"]["roles"][role_id]
                resolved["roles"][role_id] = discord.Role(data=role_data, guild=_discord.get_guild(data["guild_id"]), state=_discord._connection)
        elif x == "users":
            pass
        else:
            print("Warning: Could not resolve data of type '" + str(x) +  "'")

    return resolved

async def fetch_data(value, typ, data, _discord):
    if typ == OptionType.MEMBER:
        return await (await _discord.fetch_guild(int(data["guild_id"]))).fetch_member(int(value))
    elif typ == OptionType.CHANNEL:
        print("channel", await _discord.fetch_channel(int(value)))
        return await _discord.fetch_channel(int(value))
    elif typ == OptionType.ROLE:
        return get(await (await _discord.fetch_guild(int(data["guild_id"]))).fetch_roles(), value, lambda x: getattr(x, "id", None) == int(value))
    elif typ == 44:
        return await (await _discord.fetch_channel(data["channel_id"])).fetch_message(int(value))
    else:
        return value

def resolve_data(value, typ, data, _discord):
    resolved = resolve(data, _discord)
    if typ in [OptionType.STRING, OptionType.INTEGER, OptionType.BOOLEAN, OptionType.FLOAT]:
        return value
    elif typ == OptionType.MEMBER:
        return resolved["members"].get(value)
    elif typ == OptionType.CHANNEL:
        return resolved["channels"].get(value)
    elif typ == OptionType.ROLE:
        return resolved["roles"].get(value)
    elif typ == 44:
        return resolved["messages"].get(value)

def cache_data(value, typ, data, _discord):
    if typ in [OptionType.STRING, OptionType.INTEGER, OptionType.BOOLEAN, OptionType.FLOAT]:
        return value
    elif typ == OptionType.MEMBER:
        return _discord.get_guild(int(data["guild_id"])).get_member(int(value))
    elif typ == OptionType.CHANNEL:
        return _discord.get_channel(int(value))
    elif typ == OptionType.ROLE:
        return _discord.get_guild(int(data["guild_id"])).get_role(value)
    elif typ == 44:
        return _discord.get_guild(int(data["guild_id"])).get_partial_message(value)

async def handle_options(data, options, method, _discord: discord.Client):
    _options = {}
    for op in options:
        _options[op["name"]] = await handle_thing(op["value"], op["type"], data, _discord, method)
    return _options

async def handle_thing(value, typ, data, method, _discord):
    if method is ParseMethod.RESOLVE or method is ParseMethod.AUTO:
        try:
            return resolve_data(value, typ, data, _discord)
        except Exception as ex:
            print("Got exepction while resolving data",
                f"\n{type(ex).__name__}: {ex}\n",
                f"{__file__}:{ex.__traceback__.tb_lineno}"
            )
            if method is ParseMethod.AUTO:
                return await handle_thing(value, typ, data, ParseMethod.FETCH, _discord)
    elif method is ParseMethod.FETCH:
        try:
            return await fetch_data(value, typ, data, _discord)
        except Exception as ex:
            print("Got exepction while getting data from cache",
                f"\n{type(ex).__name__}: {ex}\n",
                f"{__file__}:{ex.__traceback__.tb_lineno}"
            )
            if method is ParseMethod.AUTO:
                return await handle_thing(value, typ, data, ParseMethod.CACHE, _discord)
    elif method is ParseMethod.CACHE:
        try:
            return cache_data(value, typ, data, _discord)
        except Exception as ex:
            print("Got exepction while resolving data",
                f"\n{type(ex).__name__}: {ex}\n",
                f"{__file__}:{ex.__traceback__.tb_lineno}"
            )
            if method is ParseMethod.AUTO:
                return await handle_thing(value, typ, data, ParseMethod.RAW, _discord)
    elif method is ParseMethod.RAW:
        return value


from .tools import MISSING

from discord import Emoji, emoji
from discord.errors import InvalidArgument

import inspect
from typing import Any, List, Union

class SelectOption():
    """
    An option for a select menu
    
    Parameters
    ----------
    value: :class:`str`
        The dev-define value of the option, max 100 characters
    label: :class:`str`
        The user-facing name of the option, max 25 characters; default \u200b ("empty" char)
    description: :class:`str`, optional
        An additional description of the option, max 50 characters
    emoji : :class:`discord.Emoji` | :class:`str`, optional
        Emoji appearing before the label; default MISSING

    Raises
    ------
    :raises: :class:`discord.InvalidArgument`: A passed argument was invalid
    """
    def __init__(self, value, label="\u200b", description=MISSING, emoji=MISSING) -> None:
        """
        Creates a new SelectOption

        Example:
        ```py
        SelectOption(label="This is a option", value="my_value", description="This is the description of the option")
        ```
        """
        self._json = {}
        
        self.label = label
        self.value = value
        if description is not MISSING:
            self.description = description
        if emoji is not MISSING:
            self.emoji = emoji
        

    @property
    def content(self) -> str:
        """
        The complete option content, consisting of the emoji and label

        :type: :class:`str`
        """
        return (self.emoji + " " if self.emoji is not None else "") + (self.label or '')
    
    @property
    def label(self) -> str:
        """
        The main text appearing on the option 

        :type: :class:`str`
        """
        return self._json["label"]
    @label.setter
    def label(self, value: str):
        if value is None:
            value = ""
        elif value is not None and type(value) is not str:
            raise InvalidArgument("label must be of type str, not " + str(type(value)))
        elif value is not None and len(value) > 100 and value > 0:
            raise InvalidArgument("label must be 100 or fewer and must be higher than 0 in length (" + str(len(value)) + ")")
        self._json["label"] = value

    @property
    def value(self) -> Any:
        """
        A unique value for the option, which will be usedd to identify the selected value

        :type: :class:`str`
        """
        return self._json["value"]
    @value.setter
    def value(self, value):
        if inspect.isclass(value):
            raise Exception()
        self._json["value"] = value

    @property
    def description(self) -> str:
        """
        A description appearing on the option

        :type: :class:`str`
        """
        return self._json.get("description")
    @description.setter
    def description(self, value):
        if value is not MISSING and type(value) is not str:
            raise InvalidArgument("description must be of type str, not " + " " + str(type(value)))
        if value is not MISSING and len(value) > 50:
            raise InvalidArgument("description must be 50 or fewer in length (" + str(len(value)) + ")")
        self._json["description"] = value
    
    @property
    def default(self) -> bool:
        """
        Whether the button is selected by default in the menu or not

        :type: :class:`bool`
        """
        return self._json.get("default", False)


    @property
    def emoji(self) -> str:
        """
        The mention of the emoji before the text

            .. note::
                For setting the emoji, you can use a str or discord.Emoji
        
        :type: :class:`str`
        """
        if "emoji" not in self._json:
            return None
        if "id" not in self._json["emoji"]:
            return self._json["emoji"]["name"]
        return f'<{"a" if "animated" in self._json["emoji"] else ""}:{self._json["emoji"]["name"]}:{self._json["emoji"]["id"]}>'
    @emoji.setter
    def emoji(self, val: Union[Emoji, str]):
        """The emoji appearing before the label"""
        if type(emoji) is str:
            self._json["emoji"] = {
                "id": None,
                "name": emoji
            }
        elif type(emoji) is Emoji:
            self._json["emoji"] = {
                "id": emoji.id,
                "name": emoji.name,
                "animated": emoji.animated
            }
        elif type(emoji) is dict:
            self._json["emoji"] = emoji


    def to_dict(self) -> dict:
        return self._json

    @classmethod
    def _fromData(cls, data) -> "SelectOption":
        """Initializes a new SelectOption from a dict
        
        Parameters
        ----------
            data: :class:`dict`
                The data to initialize from
        Returns
        -------
            :class:`~SelectOption`
                The new Option generated from the dict
        
        """
        x = SelectOption("EMPTY", "EMPTY")
        x._json = data
        return x

class SelectMenu():
    """A select menu

    Parameters
    ----------
    custom_id: :class:`str`
        The custom_id for identifying the menu, max 100 characters
    options: List[:class:`~SelectOption`]
        A list of optionns to select from
    min_values: :class:`int`, optional
        The minimum number of items that must be chosen; default ``1``, min 0, max 25
    max_values: :class:`int`, optional
        The maximum number of items that can be chosen; default ``1``, max 25
    placeholder: :class:`str`, optional
        A custom placeholder text if nothing is selected, max 100 characters; default MISSING
    default: :class:`int`, optional
        The position of the option that should be selected by default; default MISSING
    disabled: :class:`bool`, optional
        Whether the select menu should be disabled or not; default ``False``

    Raises
    ------
    :raises: :class:`discord.InvalidArgument`: A passed argument was invalid
    """
    def __init__(self, custom_id, options, min_values = 1, max_values = 1, placeholder=MISSING, default=MISSING, disabled=False) -> None:
        """
        Creates a new ui select menu

        Example:
        ```py
        SelectMenu(custom_id="my_id", options=[SelectOption(...)], min_values=2, placeholder="select something", default=0)
        ```
        """
        if type(custom_id) is not str:
            raise InvalidArgument("custom_id must be of type str, not " + str(type(custom_id)))
        if len(custom_id) > 100:
            raise InvalidArgument("custom_id must be 100 or fewer in length")
        if len(custom_id) < 1:
            raise InvalidArgument("custom_id cannot be empty")
        self._json = {
            "type": ComponentType.SELECT_MENU,
            "custom_id": custom_id,
            "disabled": disabled
        }

        if type(options) is list:
            if len(options) > 25 or len(options) == 0:
                raise InvalidArgument("You need to specify at least 1 option and at most 25 (" + str(len(options)) + ")")
            if len(options) != [x for x in options if type(x) is SelectOption]:
                self._json["options"] = [x.to_dict() for x in options]
            elif len(options) != [x for x in options if type(x) is dict]:
                self._json["options"] = options
            else:
                raise InvalidArgument("options must be of type SelectOption or dict, not " + str(type(options[0])))
        else:
            raise InvalidArgument("options must be of type List[SelectOption], not " + str(type(options)))

        if min_values is not MISSING and max_values is MISSING:
            if min_values < 0 or min_values > 25:
                raise InvalidArgument("min_values must be between 1 and 25")
            self._json["min_values"] = min_values
            self._json["max_values"] = min_values
        elif min_values is MISSING and max_values is not MISSING:
            if max_values < 0 or min_values > 25:
                raise InvalidArgument("min_values must be between 1 and 25")
            self._json["min_values"] = 1
            self._json["max_values"] = max_values
        elif min_values is not MISSING and max_values is not MISSING:
            if max_values < 0 or min_values > 25:
                raise InvalidArgument("min_values must be between 1 and 25")
            if min_values < 0 or min_values > 25:
                raise InvalidArgument("min_values must be between 1 and 25")
            self._json["min_values"] = min_values
            self._json["max_values"] = max_values

        if placeholder is not MISSING:
            self._json["placeholder"] = placeholder
        
        if default is not MISSING:
            self.set_default_option(default)

    def __str__(self) -> str:
        return self.custom_id
    
    @classmethod
    def _fromData(cls, data) -> 'SelectMenu':
        """Returns a new SelectMenu from dict data
        
        Returns
        -------
            :class:`~SelectMenu`
                The initialized SelectMenu
        
        """
        x = cls._empty()
        x._json = data
        return x

    @staticmethod
    def _empty() -> 'SelectMenu':
        """
        Creates a new "empty" select menu
        
        Returns
        -------
            :class:`~SelectMenu`
                An "empty" SelectMenu
        """
        return SelectMenu("empty", [SelectOption("EMPTY", "EMPTY", "EMPTY")], 0, 0)

    # region props
    @property
    def component_type(self) -> int:
        """
        The message component type
        
            .. note::
                The message component type will be always 3, because 3 is a select menu
        
        :type: :class:`int`
        """
        return self._json["type"]
    @property
    def custom_id(self) -> str:
        """
        The custom_id of the menu to identify it

        :type: :class:`str`
        """
        return self._json["custom_id"]
    @custom_id.setter
    def custom_id(self, value: str):
        self._json["custom_id"] = value

    @property
    def options(self) -> List[SelectOption]:
        """
        The choices in the select menu to select from

        :type: List[:class:`~SelectOption`]
        """
        return [SelectOption._fromData(x) for x in self._json["options"]]
    @options.setter
    def options(self, value: List[SelectOption]):
        self._json["options"] = [x.to_dict() for x in value]

    @property
    def default_option(self) -> Union[SelectOption, None]:
        """
        The option selected by default

        :type: :class:`~SelectOption` | :class:`None`
        """
        x = [x for x in self.options if x.default]
        if len(x) == 1:
            return x[0]
    def set_default_option(self, position: int) -> 'SelectMenu':
        """Selects the default selected option

        Parameters
        ----------
        position: :class:`int`
            The position of the option that should be default

        Raises
        ------
        :raises: :class:`discord.InvalidArgument`: A passed argument was invalid
        """
        if type(position) is not int:
            raise InvalidArgument("position has to be of type int, not " + str(type(position)))
        if position < 0 or position > len(self.options):
            raise InvalidArgument("default option position needs to be between 0 and " + str(len(self.options) - 1) + "(length of options)")
        self._json["options"][position]["default"] = True
        return self
    
    @property
    def placeholder(self) -> Union[str, None]:
        """
        Custom placeholder text if nothing is selected

        :type: :class:`str` | :class:`None`
        """
        return self._json.get("placeholder")
    @placeholder.setter
    def placeholder(self, value: str):
        self._json["placeholder"] = value

    @property
    def min_values(self) -> int:
        """
        The minimum number of items that must be chosen; default 1, min 0, max 25

        :type: :class:`int`
        """
        return self._json["min_values"]
    @min_values.setter
    def min_values(self, value: int):
        self._json["min_values"] = value

    @property
    def max_values(self) -> int:
        """
        The maximum number of items that can be chosen; default 1, max 25

        :type: :class:`int`
        """
        return self._json["max_values"]
    @max_values.setter
    def max_values(self, value: int):
        self._json["max_values"] = value
    
    @property
    def disabled(self) -> bool:
        """
        Whether the selectmenu is disabled or not

        :type: :class:`bool`
        """
        return self._json.get('disabled', False)
    @disabled.setter
    def disabled(self, value):
        self._json["disabled"] = value

    @property
    def hash(self):
        """Hash for the selectmenu

        :type: :class:`str`
        """
        return self._json.get("hash")
    # endregion

    def to_dict(self) -> dict:
        return self._json

# region Button
class Button():
    """A discord-ui button

    Parameters
    ----------
    custom_id: :class:`str`
        A identifier for the button, max 100 characters
    label: :class:`str`, optional
        Text that appears on the button, max 80 characters; default \u200b ("empty" char)
    color: :class:`str` | :class:`int`, optional
        The color of the button; default "blurple"

        .. tip:

            You can either use a string for a color or an int. Color strings are: 
            (`primary`, `blurple`), (`secondary`, `grey`), (`succes`, `green`) and (`danger`, `Red`)
            
            If you want to use integers, take a lot at the :class:`~Colors` class

    emoji: :class:`discord.Emoji` | :class:`str`, optional
        The emoji displayed before the text; default MISSING
    new_line: :class:`bool`, optional
        Whether a new line should be added before the button; default False
    disabled: :class:`bool`, optional
        Whether the button is disabled; default False

    Raises
    ------
    :raises: :class:`discord.InvalidArgument`: A passed argument was invalid
    """
    def __init__(self, custom_id, label="\u200b", color = "blurple", emoji=MISSING, new_line=False, disabled=False) -> None:
        """
        Creates a new ui-button

        Example:
        ```py
        Button("my_custom_id", "This is a cool button", "green", new_line=True)
        ```
        """
        if label is MISSING and emoji is MISSING:
            raise InvalidArgument("You need to pass a label or an emoji")
        if label is not MISSING and type(label) is not str:
            raise InvalidArgument("label must be of type str, not " + str(type(label)))
        if type(custom_id) is not str:
            raise InvalidArgument("custom_id must be of type str, not " + str(type(custom_id)))
        if type(disabled) is not bool:
            raise InvalidArgument("disabled must be of type bool") 
        if emoji is not MISSING and type(emoji) not in [Emoji, str, dict]:
            raise InvalidArgument("emoji must be of type discord.Emoji or str, not " + str(type(emoji)))
        if len(custom_id) > 100:
            raise InvalidArgument("custom_id maximum character limit (100) exceeded")
        if len(custom_id) < 1:
            raise InvalidArgument("custom_id must be longer than 0 characters")
        if label is not MISSING and len(label) > 80:
            raise InvalidArgument("lavel maximum character limit (80) exceeded")
        if label is not MISSING and len(label) < 1:
            raise InvalidArgument("label must be longer than 0 characters")
        if Colors.getColor(color) is None:
            raise InvalidArgument(str(color) + " is not a valid color")

        self.new_line = new_line
        self._json = {
            "type": ComponentType.BUTTON,
            "custom_id": custom_id,
            "style": Colors.getColor(color),
            "disabled": bool(disabled),
        }
        if label is not MISSING:
            self._json["label"] = label
        if emoji is not MISSING:
            if type(emoji) is str:
                self._json["emoji"] = {
                    "id": None,
                    "name": emoji
                }
            elif type(emoji) is Emoji:
                self._json["emoji"] = {
                    "id": emoji.id,
                    "name": emoji.name,
                    "animated": emoji.animated
                }
            elif type(emoji) is dict:
                self._json["emoji"] = emoji

    def to_dict(self):
        return self._json

    # region props
    @property
    def component_type(self) -> int:
        """
        The message component type

            .. note::
                The message component type will be always 2 (button)
        
        :type: :class:`int`
        """
        return self._json["type"]

    @property
    def content(self) -> str:
        """
        The complete content in the button ("{emoji} {label}")

        :type: :class:`str`
        """
        return (self.emoji + " " if self.emoji is not None else "") + (self.label or '')

    @property
    def custom_id(self) -> str:
        """
        The unique custom_id for identifiying the button

        :type: :class:`str`
        """
        return self._json["custom_id"]
    @custom_id.setter
    def custom_id(self, val: str):
        if type(val) is not str:
            raise InvalidArgument("custom_id must be of type str, not " + str(type(val)))
        
        if len(val) > 100:
            raise InvalidArgument("custom_id must be shorter than 100 characters")
        if len(val) < 1:
            raise InvalidArgument("custom_id must be longer than 0 characters")

        self._json["custom_id"] = str(val)

    @property
    def label(self) -> str:
        """
        The label displayed on the button

        :type: :class:`str`
        """
        return self._json.get("label", None)
    @label.setter
    def label(self, val: str):
        if val is None:
            val = ""
        elif val is not None and type(val) is not str:
            raise InvalidArgument("label must be of type str, not " + str(type(val)))
        elif val is not None and len(val) > 80:
            raise InvalidArgument("label must be shorter than 80 characters")
        elif val is not None and len(val) < 1:
            raise InvalidArgument("label must be longer than 0 characters")

        self._json["label"] = str(val)

    @property
    def color(self) -> int:
        """
        The color for the button

        :type: :class:`int`, one of :class:`~ButtonStyles`
        """
        return self._json["style"]
    @color.setter
    def color(self, val):
        if Colors.getColor(val) is None:
            raise InvalidArgument(str(val) + " is not a valid color")
        self._json["style"] = Colors.getColor(val)
    
    @property
    def emoji(self) -> str:
        """The mention of the emoji before the text
        
            .. note::
                For setting the emoji, you can use a str or discord.Emoji          
        
        :type: :class:`str`
        """
        if "emoji" not in self._json:
            return None
        if "id" not in self._json["emoji"]:
            return self._json["emoji"]["name"]
        return f'<{"a" if "animated" in self._json["emoji"] else ""}:{self._json["emoji"]["name"]}:{self._json["emoji"]["id"]}>'
    @emoji.setter
    def emoji(self, val: Union[Emoji, str]):
        if type(val) not in [Emoji, str]:
            raise InvalidArgument("emoji msut be of type discord.Emoji or str, not " + str(type(val)))
        if type(val) is str:
            self._json["emoji"] = {
                "id": None,
                "name": val
            }
        elif type(val) is Emoji:
            self._json["emoji"] = {
                "id": val.id,
                "name": val.name,
                "animated": val.animated
            }

    @property
    def disabled(self) -> bool:
        """
        Whether the button is dissabled
        If true, the button is shown but not clickable

        :type: :class:`bool`
        """
        return self._json["disabled"] if "disabled" in self._json else False
    @disabled.setter
    def disabled(self, val):
        if type(val) != bool:
            raise InvalidArgument("disabled must be type of bool")
        if "disabled" in self._json:
            self._json["disabled"] = bool(val)
        else:
            self._json |= {"disabled": bool(val)}
    
    @property
    def hash(self) -> str:
        """
        The calculated hash from the discord api for the button

        :type: :class:`str`
        """
        return self._json.get('hash', None)
    # endregion
    
    @classmethod
    def _empty(cls):
        """Returns an "empty" button"""
        return cls("empty", "empty")
    @classmethod
    def _fromData(cls, data, new_line=False) -> 'Button':
        """Returns a new button initialized from api response data

        Returns
        -------
        Button
            The initialized button
        """
        b = cls("empty", "empty")
        b._json = data
        b.new_line = new_line
        return b

class LinkButton():
    """
    A discord-ui linkbutton

    Parameters
    ----------
    url: :class:`str`
        A url which will be opened when pressing the button
    label: :class:`str`, optional
        Text that appears on the button, max 80 characters; default \u200b ("empty" char)
    emoji: :class:`discord.Emoji` | :class:`str`, optional
        Emoji that appears before the label; default MISSING
    new_line: :class:`bool`, optional
        Whether a new line should be added before the button; default False
    disabled: :class:`bool`, optional
        Whether the button is disabled; default False

    
    Raises
    ------
    :raises: :class:`discord.InvalidArgument`: A passed argument was invalid
    """
    def __init__(self, url, label="\u200b", emoji=MISSING, new_line=False, disabled=False) -> None:
        """
        Creates a new LinkButton object
        
        Example:
        ```py
        LinkButton("https://discord.com", "press me (if you can)!", emoji="😀", disabled=True)
        ```
        """
        if label is MISSING and emoji is MISSING:
            raise InvalidArgument("You need to pass label or emoji")
        if label is not MISSING and type(label) is MISSING:
            raise InvalidArgument("label must be of type str, not " + str(type(label)))
        if label is not MISSING and type(label) is not str:
            raise InvalidArgument("label must be of type str, not " + str(type(label)))
        if type(url) is not str:
            raise InvalidArgument("url must be of type str, not " + str(type(url)))
        if not url.startswith("http://") and not url.startswith("https://"):
            raise InvalidArgument("Link must start with 'http://' or 'https://'")
        if type(disabled) is not bool:
            raise InvalidArgument("disabled must be of type bool, not " + str(type(disabled)))
        if emoji is not MISSING and type(emoji) not in [Emoji, str, dict]:
            raise InvalidArgument("emoji msut be of type discord.Emoji or str, not " + str(type(emoji)))
        if label is not MISSING and len(label) > 80:
            raise InvalidArgument("lavel maximum character limit (80) exceeded")
        if label is not MISSING and len(label) < 1:
            raise InvalidArgument("label must be longer than 0 characters")

        self.new_line = new_line
        self._json = {
            "type": ComponentType.BUTTON,
            "url": url,
            "style": 5,
            "disabled": disabled
        }
        if label is not MISSING:
            self._json["label"] = label
        if emoji is not MISSING:
            if type(emoji) is str:
                self._json["emoji"] = {
                    "id": None,
                    "name": emoji
                }
            elif type(emoji) is Emoji:
                self._json["emoji"] = {
                    "id": emoji.id,
                    "name": emoji.name,
                    "animated": emoji.animated
                }
            elif type(emoji) is dict:
                self._json["emoji"] = emoji

    def to_dict(self):
        return self._json

    # region props
    @property
    def component_type(self) -> int:
        """The message component type for this component

            .. note::
                The component type will always be 2 (button)
        
        :type: :class:`int`
        """
        return self._json["type"]

    @property
    def content(self) -> str:
        """
        The whole content of the button text, consiting of the emoji and the label

        :type: :class:`str`
        """
        return (self.emoji + " " if self.emoji is not None else "") + (self.label or '')

    @property
    def url(self) -> str:
        """
        The url which will be opened after the button is pressed

        :type: class:`str`
        """
        return self._json["url"]
    @url.setter
    def url(self, val: str):
        if type(val) is not str:
            raise InvalidArgument("url must be of type str, not " + str(type(val)))
        if not val.startswith("http://") and not val.startswith("https://"):
            raise InvalidArgument("Link must start with 'https://' or 'http://'")
        self._json["url"] = val
    
    @property
    def label(self) -> str:
        """
        The text that appears on the button

        :type: :class:`str`
        """
        return self._json.get("label", None)
    @label.setter
    def label(self, val):
        if type(val) is not str:
            raise InvalidArgument("label must be of type str, not " + str(type(val)))
        self._json["label"] = val 

    @property
    def emoji(self) -> str:
        """The mention of the emoji before the text

            .. note::
                For setting the emoji you can use `str` or `discord.Emoji`

        :type: :class:`str`
        """
        if "emoji" not in self._json:
            return None
        if "id" not in self._json["emoji"]:
            return self._json["emoji"]["name"]
        return f'<{"a" if "animated" in self._json["emoji"] else ""}:{self._json["emoji"]["name"]}:{self._json["emoji"]["id"]}>'
    @emoji.setter
    def emoji(self, val: Union[Emoji, str]):
        if type(val) not in [Emoji, str]:
            raise InvalidArgument("emoji msut be of type discord.Emoji or str, not " + str(type(val)))
        if type(val) is str:
            self._json["emoji"] = {
                "id": None,
                "name": val
            }
        elif type(val) is Emoji:
            self._json["emoji"] = {
                "id": val.id,
                "name": val.name,
                "animated": val.animated
            }

    @property
    def color(self) -> int:
        """
        The color sttyle for the button
        
            .. note::
                The color style Will always be 5 (link_button)
        
        :type: :class:`int`
        """
        return self._json["style"]

    @property
    def disabled(self) -> bool:
        """
        Whether the button is disabled or not
        
        :type: :class:`bool`
        """
        return self._json["disabled"]
    @disabled.setter
    def disabled(self, val):
        self._json["disabled"] = val
    # endregion

    @classmethod
    def _empty(cls):
        """Returns an empty button"""
        return cls("empty", "empty")
    @classmethod
    def _fromData(cls, data, new_line=False) -> 'LinkButton':
        """Returns a new Linkbutton initialized from api response data"""
        b = cls("https://empty", "empty")
        b._json = data
        b.new_line = new_line
        return b


class Colors:
    """
    A list of button styles (colors) in message components
    """
    Primary     =   blurple         = 1
    Secondary   =   grey            = 2
    Succes      =   green           = 3
    Danger      =   red             = 4
    URL                             = 5

    @classmethod
    def getColor(cls, s):
        if type(s) is int:
            return s
        s = s.lower()
        if s in ("blurple", "primary"):
            return cls.blurple
        if s in ("grey", "gray", "secondary"):
            return cls.grey
        if s in ("green", "succes"):
            return cls.green
        if s in ("red", "danger"):
            return cls.red
# endregion


class ActionRow():
    """Alternative to setting ``new_line`` in a full component list or putting the components in a list 
    
    Only works for :class:`~Button` and :class:`~LinkButton`, because :class:`~SelectMenu` is always in a new line
    
    Parameters
    ----------
        disbaled: :class:`bool`, optional
            Whether all components should be disabled; default False   
    """
    def __init__(self, *items, disbaled = False):
        """Creates a new component list

        Examples
        ```py
        ActionRow(Button(...), Button(...))
        
        ActionRow([Button(...), Button(...)])
        ```
        """
        self.items = items[0] if all(type(i) is list for i in items) else items
        """The componetns in the action row"""
        self.component_type = 1
        
    def disable(self, disable=True) -> 'ActionRow':
        for i, _ in enumerate(self.items):
            self.items[i].disabled = disable
        return self
    def filter(self, check = lambda x: ...):
        """Filters all components
        
        Parameters
        ----------
            check: :class:`lambda`
                What condition has to be True that the component will pass the filter
        Returns
        -------
            :returns: The filtered components
            :type: List[:class:`~Button` | :class:`~LinkButton`]
        
        """
        return [x for x in self.items if check(x)]


class ComponentType:
    """
    A list of component types
    """
    ACTION_ROW = 1
    BUTTON = 2
    SELECT_MENU = 3



from discord.state import ConnectionState
from .slash.types import ContextCommand, SlashCommand, SubSlashCommand
from .tools import MISSING
from .http import BetterRoute, jsonifyMessage, send_files
from .components import ActionRow, Button, LinkButton, SelectMenu, SelectOption

import discord
from discord.ext import commands as com
from discord.errors import HTTPException, InvalidArgument
import typing


class Interaction():
    def __init__(self, application_id, state, data, user=MISSING) -> None:
        self._application_id = application_id
        self._state: ConnectionState = state

        self._deferred = False
        self._deferred_hidden = False
        self._responded = False

        if user is not MISSING:
            self.member = user
            """The user who created the interaction
            
            :type: :class:`discord.Member`
            """
        self.interaction = {
            "token": data["token"],
            "id": data["id"]
        }


    async def defer(self, hidden=False):
        """This will acknowledge the interaction. This will show the (*Bot* is thinking...) Dialog

        .. note::
            
            This function should be used if the bot needs more than 15 seconds to respond
        
        Parameters
        ----------
            hidden: :class:`bool`, optional
                Whether the loading thing should be only visible to the user; default False.
        
        """
        if self._deferred:
            raise Exception("interaction was already deferred")
        body = {"type": 5}
        
        if hidden is True:
            body["data"] = {"flags": 64}
            self._deferred_hidden = True
        
        await self._state.http.request(BetterRoute("POST", f'/interactions/{self.interaction["id"]}/{self.interaction["token"]}/callback'), json=body)
        self._deferred = True

    async def respond(self, content=MISSING, *, tts=False, embed=MISSING, embeds=MISSING, file=MISSING, files=MISSING, nonce=MISSING,
    allowed_mentions=MISSING, mention_author=MISSING, components=MISSING, delete_after=MISSING, hidden=False,
    ninja_mode=False) -> typing.Union['Message', 'EphemeralMessage']:
        """Responds to the interaction
        
        Parameters
        ----------
        content: :class:`str`, optional
            The raw message content
        tts: :class:`bool` 
            Whether the message should be send with text-to-speech
        embed: :class:`discord.Embed`
            Embed rich content
        embeds: List[:class:`discord.Embed`]
            A list of embeds for the message
        file: :class:`discord.File`
            The file which will be attached to the message
        files: List[:class:`discord.File`]
            A list of files which will be attached to the message
        nonce: :class:`int`
            The nonce to use for sending this message
        allowed_mentions: :class:`discord.AllowedMentions`
            Controls the mentions being processed in this message
        mention_author: :class:`bool`
            Whether the author should be mentioned
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
            A list of message components to be included
        delete_after: :class:`float`
            After how many seconds the message should be deleted, only works for non-hiddend messages; default MISSING
        hidden: :class:`bool`
            Whether the response should be visible only to the user 
        ninja_mode: :class:`bool`
            If true, the client will respond to the button interaction with almost nothing and returns nothing
        
        Returns
        -------
        :return: Returns the sent message
        :type: :class:`~Message` | :class:`EphemeralMessage`

            .. note::
                
                If the response is hidden, a EphemeralMessage will be returned, which is an empty class

        """
        
        if ninja_mode:
            try:
                await self._state.http.request(BetterRoute("POST", f'/interactions/{self.interaction["id"]}/{self.interaction["token"]}/callback'), json={
                    "type": 6
                })
            except HTTPException as x:
                if "value must be one of (4, 5)" in str(x).lower():
                    print(x, "The 'ninja_mode' parameter is not supported for slash commands!")
                else:
                    print(x)
            finally:
                return

        if self._responded:
            await self.send(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, components=components, hidden=hidden)
            return

        
        payload = jsonifyMessage(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, components=components)
        
        hide_message = self._deferred_hidden or not self._deferred and hidden 

        if delete_after is not MISSING and hide_message is True:
            raise InvalidArgument("Cannot delete hidden message (delete_after won't work)")

        if hide_message:
            payload["flags"] = 64

        if (file is not MISSING or files is not MISSING) and self._deferred is False or (hide_message is True and not self._deferred):
            await self.defer(hidden=hide_message)
                
        if not self._deferred:
            route = BetterRoute("POST", f'/interactions/{self.interaction["id"]}/{self.interaction["token"]}/callback')
            await self._state.http.request(route, json={
                    "type": 4,
                    "data": payload
                })
        else:
            route = BetterRoute("PATCH", f'/webhooks/{self._application_id}/{self.interaction["token"]}/messages/@original')
            r = await self._state.http.request(route, json=payload)
            if hide_message:
                self._responded = True
                return EphemeralMessage(state=self._state, channel=self._state.get_channel(int(r["channel_id"])), data=r)

        if file is not MISSING and files is not MISSING:
            await send_files(route=BetterRoute("PATCH", f"/webhooks/{self._application_id}/{self.interaction['token']}/messages/@original"), 
                files=[file] if files is MISSING else files, payload=payload, http=self._state.http)
        
        if not hide_message:
            responseMSG = await self._state.http.request(BetterRoute("GET", f"/webhooks/{self._application_id}/{self.interaction['token']}/messages/@original"))
            msg = await getResponseMessage(self._state, data=responseMSG, application_id=self._application_id, response=False)
            if delete_after is not MISSING:
                await msg.delete(delete_after)
            return msg

    async def send(self, content=None, *, tts=False, embed=MISSING, embeds=MISSING, file=MISSING, files=MISSING, nonce=MISSING,
    allowed_mentions=MISSING, mention_author=MISSING, components=MISSING, hidden=False) -> typing.Union['Message', 'EphemeralMessage']:
        """Sends a message to the interaction using a webhook
        
        Parameters
        ----------
        content: :class:`str`, optional
            The raw message content
        tts: `bool` 
            Whether the message should be send with text-to-speech
        embed: :class:`discord.Embed`
            Embed rich content
        embeds: List[:class:`discord.Embed`]
            A list of embeds for the message
        file: :class:`discord.File`
            The file which will be attached to the message
        files: List[:class:`discord.File`]
            A list of files which will be attached to the message
        nonce: :class:`int`
            The nonce to use for sending this message
        allowed_mentions: :class:`discord.AllowedMentions`
            Controls the mentions being processed in this message
        mention_author: :class:`bool`
            Whether the author should be mentioned
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
            A list of message components to be included
        hidden: :class:`bool`
            Whether the response should be visible only to the user 
        ninja_mode: :class:`bool`
            If true, the client will respond to the button interaction with almost nothing and returns nothing
        
        Returns
        -------
        :return: Returns the sent message
        :type: :class:`~Message` | :class:`EphemeralMessage`

            .. note::
                If the response is hidden, a EphemeralMessage will be returned, which is an empty class
        """
        if self._responded is False:
            return await self.respond(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, components=components, hidden=hidden)

        payload = jsonifyMessage(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, components=components)
        
        if hidden:
            payload["flags"] = 64

        route = BetterRoute("POST", f'/webhooks/{self._application_id}/{self.interaction["token"]}')
        r = await self._state.http.request(route, json=payload)

        if file is not MISSING and files is not MISSING:
            await send_files(route=BetterRoute("PATCH", f"/webhooks/{self._application_id}/{self.interaction['token']}/messages/@original"), 
                files=[file] if files is MISSING else files, payload=payload, http=self._state.http)

        if hidden:
            return EphemeralMessage(state=self._state, channel=self._state.get_channel(int(r["channel_id"])), data=r)

        return await getResponseMessage(self._state, r, response=False)

        

class SelectedMenu(Interaction, SelectMenu):
    """A :class:`~SelectMenu` object in which an item was selected"""
    def __init__(self, data, application_id, user, s, state) -> None:
        Interaction.__init__(self, application_id, state, data)
        SelectMenu.__init__(self, "EMPTY", [SelectOption("EMPTY", "EMPTY")], 0, 0)
        self._json = s.to_dict()
        self.values: typing.List[SelectOption] = []
        """The list of values which were selected
        
        :type: :class:`~SelectOption`
        """
        
        for val in data["data"]["values"]:
            for x in self.options:
                if x.value == val:
                    self.values.append(x)

        self.member: discord.Member = user
        """The member who selected the value"""

class PressedButton(Interaction, Button):
    """A :class:`~Button` object that was pressed"""
    def __init__(self, data, application_id, user, b, state) -> None:
        Interaction.__init__(self, application_id, state, data)
        Button.__init__(self, "empty", "empty")
        self._json = b.to_dict()

        """interaction: :class:`dict`
        
        The most important stuff from the received interaction
        
        *  ``token``
                The interaction token
        *   ``id``
                The ID for the interaction
        """
        self.member: discord.Member = user
        """The user who pressed the button"""

class SlashedCommand(Interaction, SlashCommand):
    """A :class:`~SlashCommand` object that was used"""
    def __init__(self, client, command: SlashCommand, data, user, channel, guild_ids = None) -> None:
        Interaction.__init__(self, client.user.id, client._connection, data, user)
        SlashCommand.__init__(self, None, "EMPTY", guild_ids)
        self._json = command.to_dict()
        self.member: discord.Member = user
        """The user who created the interaciton"""
        self.channel: discord.TextChannel = channel
        """The channel where the slash command was used"""
        self.guild_ids = guild_ids

class SlashedSubCommand(SlashedCommand, SubSlashCommand):
    """A Sub-:class:`~SlashCommand` command that was used"""
    def __init__(self, client, command, data, user, channel, guild_ids) -> None:
        SlashedCommand.__init__(self, client, command, data, user, channel, guild_ids=guild_ids)
        SubSlashCommand.__init__(self, None, "EMPTY", "EMPTY")


class SlashedContext(Interaction, ContextCommand):
    def __init__(self, client, command: ContextCommand, data, user, channel, guild_ids = None) -> None:
        Interaction.__init__(self, client.user.id, client._connection, data, user)
        ContextCommand.__init__(self)
        self._json = command.to_dict()
        self.member: discord.Member = user
        self.channel: discord.TextChannel = channel
        self.guild_ids = guild_ids


async def getResponseMessage(state: ConnectionState, data, application_id = None, user=None, response = True):
    """
    Async function to get the response message

    Parameters
    -----------------

    state: :class:`discord.state.ConnectionState`
        The discord bot client
    data: :class:`dict`
        The raw data
    user: :class:`discord.User`
        The User which pressed the button
    response: :class:`bool`
        Whether the Message returned should be of type `ResponseMessage` or `Message`

    Returns
    -------
    :class:`~Message` | :class:`~ResponseMessage`
        The sent message

    .. note::
            If the message comes from an interaction, it will be of type :class:`~ResponseMessage`, if it is sent to a textchannel, it will be of type :class:`~Message`
    """
    channel = state.get_channel(int(data["channel_id"]))
    if response and user:
        if data["message"]["flags"] == 64:
            return EphemeralMessage(state=state, channel=channel, data=data["message"])
        return ResponseMessage(state=state, application_id=application_id, channel=channel, data=data, user=user)

    return Message(state=state, channel=channel, data=data)

class Message(discord.Message):
    """A fixed :class:`discord.Message` optimized for components"""
    def __init__(self, *, state, channel, data):
        self.__slots__ = discord.Message.__slots__ + ("components", "supressed")
    
        super().__init__(state=state, channel=channel, data=data)
        self.components: typing.List[typing.Union[Button, LinkButton, SelectMenu]] = []
        """The components in the message
        
        :type: List[]:class:`~Button` | :class:`~LinkButton` | :class:`SelectMenu`]
        """
        self.suppressed = False
        
        self._update_components(data)

    # region attributes
    @property
    def buttons(self):
        """The button components in the message
        
        :type: List[:class:`~Button` | :class:`~LinkButton`]
        """
        if hasattr(self, "components") and self.components is not None:
            return [x for x in self.components if type(x) in [Button, LinkButton]]
        return []
    @property
    def select_menus(self):
        """The select menus components in the message

        :type: List[:class:`~SelectMenu`]
        """
        if hasattr(self, "components") and self.components is not None:
            return [x for x in self.components if type(x) is SelectMenu]
        return []
    # endregion

    def _update_components(self, data):
        """Updates the message components"""
        if data.get("components") is None:
            return
        if len(data["components"]) == 0:
            self.components = []
        elif len(data["components"]) > 1:
            # multiple lines
            for componentWrapper in data["components"]:
                # newline
                for index, com in enumerate(componentWrapper["components"]):
                    if com["type"] == 2:
                        self.components.append(
                            Button._fromData(com, index == 0)
                                if "url" not in com else 
                            LinkButton._fromData(com, index == 0)
                        )
                    elif com["type"] == 3:
                        self.components.append(
                            SelectMenu._fromData(com)
                        )
        elif len(data["components"][0]["components"]) > 1:
            # All inline
            for index, com in enumerate(data["components"][0]["components"]):
                if com["type"] == 2:
                    self.components.append(Button._fromData(com, index == 0) if "url" not in com else LinkButton._fromData(com, index == 0))
                elif com["type"] == 3:
                    self.components.append(SelectedMenu._fromData(com))
        else:
            # One button
            component_type = int(data["components"][0]["components"][0]["type"])
            component = data["components"][0]["components"][0]

            if component_type == 2:
                self.components.append(Button._fromData(component) if "url" not in component else LinkButton._fromData(component))
            elif component_type == 3:
                self.components.append(SelectedMenu._fromData(component))
            else:
                print("unknown component type")

    def _update(self, data):
        super()._update(data)
        self._update_components(data)

    async def edit(self, *, content=MISSING, embed=MISSING, embeds=MISSING, attachments=MISSING, suppress=MISSING, 
        delete_after=MISSING, allowed_mentions=MISSING, components=MISSING):
        """Edits the message and updates its properties

        .. note::

            If a paremeter is `None`, the attribute will be removed from the message

        Parameters
        ----------------
        content: :class:`str`
            The new message content
        embed: :class:`discord.Embed`
            Embed rich content
        embeds: List[:class:`discord.Embed`]
            The new list of discord embeds
        attachments: List[:class:`discord.Attachment`]
            A list of new attachments
        supress: :class:`bool`
            Whether the embeds should be shown
        delete_after: :class:`float`
            After how many seconds the message should be deleted
        allowed_mentions: :class:`discord.AllowedMentions`
            The mentions proceeded in the message
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
            A list of components to be included the message
        """
        payload = jsonifyMessage(content, embed=embed, embeds=embeds, allowed_mentions=allowed_mentions, attachments=attachments, suppress=suppress, flags=self.flags.value, components=components)
        if suppress:
            self.suppressed = suppress
        data = await self._state.http.edit_message(self.channel.id, self.id, **payload)
        self._update(data)

        if delete_after is not MISSING:
            await self.delete(delay=delete_after)

    async def disable_action_row(self, row, disable = True):
        """Disables an action row of components in the message
        
        Parameters
        ----------
            row: :class:`int` |  :class:`range`
                Which rows to disable, first row is ``0``; If row parameter is of type :class:`int`, the nth row will be disabled, if type is :class:`range`, the range is going to be iterated and all rows in the range will be disabled

            disable: :class:`bool`, optional
                Whether to disable (``True``) or enable (``False``) the components; default True

        Raises
        ------
            :raises: :class:`InvalidArgument` : The specified range was out of the possible range of the component rows 
            :raises: :class:`InvalidArgument` : The specified row was out of the possible range of the component rows 
        
        """
        comps = []
        if type(row) is range:
            for i, _ in enumerate(self.action_rows):
                if i > len(self.action_rows) - 1 or i < 0:
                    raise InvalidArgument("the specified range is invalid! It has to be between 0 and " + str(len(self.action_rows) - 1))
                for comp in self.action_rows[i]:
                    if i in row:
                        comp.disabled = disable
                    comps.append(comp)
        else:
            for i, _ in enumerate(self.action_rows):
                if i > len(self.action_rows) - 1 or i < 0:
                    raise InvalidArgument("the specified row is invalid! It has to be between 0 and " + str(len(self.action_rows) - 1))
                for comp in self.action_rows[i]:
                    if i == row:
                        comp.disabled = disable
                    comps.append(comp)
        await self.edit(components=comps)

    async def disable_components(self, disable = True):
        """Disables all component in the message
        
        Parameters
        ----------
            disable: :class:`bool`, optional
                Whether to disable (``True``) or enable (``False``) als components; default True
        
        """
        fixed = []
        for x in self.components:
            x.disabled = disable
            fixed.append(x)
        await self.edit(components=fixed)

    @property
    def action_rows(self):
        """The action rows in the message

        :type: List[:class:`~Button` | :class:`LinkButton` | :class:`SelectMenu`]
        """
        rows: typing.List[typing.List[typing.Union[Button, LinkButton, SelectMenu]]] = []

        c_row = []
        i = 0
        for x in self.components:
            if getattr(x, 'new_line', True) == True and i > 0:
                rows.append(ActionRow(c_row))
                c_row = []
            c_row.append(x)
            i += 1
        if len(c_row) > 0:
            rows.append(c_row) 
        return rows

    @typing.overload
    async def wait_for(self, client, event_name="button") -> PressedButton: ...
    @typing.overload 
    async def wait_for(self, client, event_name="select") -> SelectedMenu: ...
    @typing.overload
    async def wait_for(self, client, event_name, custom_id) -> PressedButton: ...
    @typing.overload
    async def wait_for(self, client, event_name, custom_id) -> SelectedMenu: ...
    @typing.overload
    async def wait_for(self, client, event_name, timeout) -> PressedButton: ...
    @typing.overload
    async def wait_for(self, client, event_name, timeout) -> SelectedMenu: ...
    async def wait_for(self, client, event_name: typing.Literal["select", "button"], custom_id=MISSING, timeout=MISSING) -> typing.Union[PressedButton, SelectedMenu]:
        """Waits for a message component to be invoked in this message

        Parameters
        -----------
        client: :class:`discord.ext.commands.Bot`
            The discord client
        event_name: :class:`str`
            The name of the event which will be awaited [``"select"`` | ``"button"``] 

            .. note::

                The event_name must be ``select`` for a select menu selection and ``button`` for a button press
        
        custom_id: :class:`str`, Optional
            Filters the waiting for a custom_id
        
        timeout: :class:`float`, Optional
            After how many seconds the waiting should be canceled. 
            Throws an :class:`asyncio.TimeoutError` Exception

        Raises
        ------
            :raises: :class:`InvalidArgument` : The event name passed was invalid 

        Returns
        ----------
        :returns: The component that was waited for
        :type: :class:`~PressedButton` | :class:`~SelectedMenu`
        """
        if event_name.lower() in ["button", "select"]:
            def check(btn, msg):
                if msg.id == self.id:
                    if custom_id is not MISSING and btn.custom_id == custom_id:
                        return True
                    return True

            return (await client.wait_for('button_press' if event_name.lower() == "button" else "menu_select", check=check, timeout=timeout))[0]
        
        raise InvalidArgument("Invalid event name, event must be 'button' or 'select', not " + str(event_name))


class ResponseMessage(Interaction, Message):
    r"""A message Object which extends the `Message` Object optimized for an interaction component"""
    def __init__(self, *, state, application_id, channel, data, user):
        Interaction.__init__(self, application_id, state, data)
        Message.__init__(self, state=state, channel=channel, data=data["message"])

        self.interaction_component = None

        if int(data["data"]["component_type"]) == 2:
            for x in self.buttons:
                if hasattr(x, 'custom_id') and x.custom_id == data["data"]["custom_id"]:
                    self.interaction_component = PressedButton(data, self._application_id, user, x, self._state)
        elif int(data["data"]["component_type"]) == 3:
            for x in self.select_menus:
                if x.custom_id == data["data"]["custom_id"]:
                    self.interaction_component = SelectedMenu(data, self._application_id, user, x, self._state)

class WebhookMessage(Message, discord.WebhookMessage):
    def __init__(self, *, state, channel, data):
        Message.__init__(self, state=state, channel=channel, data=data)
        discord.WebhookMessage.__init__(self, state=state, channel=channel, data=data)
    async def edit(self, **fields):
        """Edits the message

        content: :class:`str`, Optional
            The content to edit the message with or None to clear it.
        embed: :class:`discord.Embed`
            Embed rich content
        embeds: List[:class:`discord.Embed`], Optional
            A list of embeds to edit the message with.
        supress: :class:`bool`
            Whether to supress all embeds in the message
        allowed_mentions: :class`discord.AllowedMentions`
            Controls the mentions being processed in this message. See `discord.abc.Messageable.send` for more information.
        """
        return await self._state._webhook._adapter.edit_webhook_message(message_id=self.id, payload=jsonifyMessage(**fields))
    

class EphemeralComponent(Interaction):
    """Represents a component in a hidden message
    
    .. note::
    
        You will only get this class for components if you set a component on a hidden response in an interaction and the component was used
    
    This class only has the custom_id of the used component, the component type and values
    """
    def __init__(self, application_id, state, user, data) -> None:
        Interaction.__init__(self, application_id, state, data, user)
        self.custom_id = data["data"]["custom_id"]
        self.component_type = data["data"]["component_type"]
        if self.component_type == 3:
            class EphemeralValue():
                def __init__(self, value) -> None:
                    self.name = None
                    self.value = value
            self.values = [EphemeralValue(x) for x in data["data"]["values"]]


class EphemeralMessage(Message):
    """Represents a hidden (ephemeral) message
    
    .. note::
        
        You will get this class for a message only if you sent a hidden response

    This class is almost the same as the :class:`~Message` class, but you can't edit, delete or reply to the message
    
    If you want to "reply" to id, use the `interaction.send` method instead
    """
    def __init__(self, *, state, channel, data):
        if data.get("attachments") is None:
            data["attachments"] = []
        if data.get("embeds") is None:
            data["embeds"] = []
        if data.get("edited_timestamp") is None:
            data["edited_timestamp"] = None
        if data.get("type") is None:
            data["type"] = 0
        if data.get("pinned") is None:
            data["pinned"] = False
        if data.get("mention_everyone") is None:
            data["mention_everyone"] = False
        if data.get("tts") is None:
            data["tts"] = False
        if data.get("content") is None:
            data["content"] = None
        Message.__init__(self, state=state, channel=channel, data=data)
    async def edit(self, **kwargs):
        raise Exception("Cannot edit ephemeral message")
    async def delete(self, *, delay):
        raise Exception("Cannot delete ephemeral message")
    async def respond(self, **kwargs):
        raise Exception("Cannot reply to ephemeral message. Instead, use the '.send' function of the interaction")
    



from .slash.tools import ParseMethod, handle_options, handle_thing, resolve
from .slash.http import create_global_command, create_guild_command, delete_global_command, delete_guild_command, delete_guild_commands, edit_global_command, edit_guild_command, get_command, get_global_commands, get_guild_commands, delete_global_commands, get_id
from .slash.types import MessageCommand, OptionType, SlashCommand, SlashOption, SlashPermission, SubSlashCommand, SubSlashCommandGroup, UserCommand
from .tools import MISSING, get, get_index
from .http import jsonifyMessage, BetterRoute, send_files
from .receive import EphemeralComponent, Message, SlashedContext, WebhookMessage, SlashedCommand, SlashedSubCommand, getResponseMessage

import discord
from discord.errors import Forbidden, InvalidArgument
from discord.ext import commands as com

import asyncio
from typing import Dict, List, Tuple, Union

class Slash():
    """
    A class for using slash commands
    
    Parameters
    ----------
        client: :class:`commands.Bot`
            The bot client

        parse_method: :class:`bool`, optional
            How received option data should be treated; Default ``ParseMethod.AUTO``

        delete_unused: :class:`bool`, optional
            Whether the commands that are not registered by this slash extension should be deleted in the api; Default ``False``

        wait_sync: :class:`float`, optional
            How many seconds will be waited until the commands are going to be synchronized; Default ``1``


    Example
    ------------------
    Example for using slash commands

    .. code-block::

        ...
        # Your bot declaration and everything
        slash = Slash(client)

    For creating a slash command use

    .. code-block::

        ...
        @slash.command(name="my_command", description="this is my slash command", options=[SlashOption(str, "option", "this is an option")])
        async def command(ctx: SlashedCommand):
            ...
    
    For subcommands use

    .. code-block::

        ...
        @slash.subcommand(base_name="base", name="sub", description="this is a sub command")
        async def subcommand(ctx: SlashedSubCommand):
            ...

    And for subcommand groups use

    .. code-block::

        ...
        @slash.subcommand_group(base_names=["base", "group"], name="sub", description="this is a sub command group")
        async def subcommand(ctx: SlasedSubCommand):
            ...
        

    """
    def __init__(self, client, parse_method = ParseMethod.AUTO, delete_unused = False, wait_sync = 1) -> None:
        """Creates a new slash command thing
        
        Example
        ```py
        Slash(client)
        ```
        """
        self.parse_method: int = parse_method
        self.delete_unused: bool = delete_unused
        self.wait_sync: float = wait_sync

        self._discord: com.Bot = client
        self.commands: Dict[(str, SlashCommand)] = {}
        self.subcommands: Dict[(str, Dict[(str, SubSlashCommand)])] = {}
        self.subcommand_groups: Dict[(str, Dict[(str, SubSlashCommandGroup)])] = {}
        self.context_commands = {"message": {}, "user": {}}
        self._discord.add_listener(self._on_socket_response, 'on_socket_response')

        self.ready = False
        async def client_ready():
            await asyncio.sleep(self.wait_sync or 1)
            self._discord.loop.create_task(self.add_commands())
            self.ready = True
        self._discord.add_listener(client_ready, "on_ready")

    async def _on_socket_response(self, msg):
        """Will be executed if the bot receives a socket response"""
        if msg["t"] != "INTERACTION_CREATE":
            return
        data = msg["d"]

        if int(data["type"]) not in [1, 2]:
            return

        guild = await self._discord.fetch_guild(data["guild_id"])
        user = discord.Member(data=data["member"], guild=guild, state=self._discord._connection)
        channel = await self._discord.fetch_channel(data["channel_id"])

        if data["data"]["type"] == 1:
            x = self.commands.get(data["data"]["name"])
            if x:
                options = {}
                if data["data"].get("options") is not None:
                    options = await handle_options(data, data["data"]["options"], self.parse_method, self._discord)
                await x.callback(SlashedCommand(self._discord, command=x, data=data, user=user, channel=channel, guild_ids=x.guild_ids), **options)
                return
        elif data["data"]["type"] == 2:
            x = self.context_commands["user"].get(data["data"]["name"])
            if x:
                member = await handle_thing(data["data"]["target_id"], OptionType.MEMBER, data, self.parse_method, self._discord)
                await x.callback(SlashedContext(self._discord, command=x, data=data, user=user, channel=channel, guild_ids=x.guild_ids), member)
        elif data["data"]["type"] == 3:
            x = self.context_commands["message"].get(data["data"]["name"])
            if x:
                message = await handle_thing(data["data"]["target_id"], 44, data, self.parse_method, self._discord)
                await x.callback(SlashedContext(self._discord, command=x, data=data, user=user, channel=channel, guild_ids=x.guild_ids), message)
    
        fixed_options = []
        x_base = self.subcommands.get(data["data"]["name"]) or self.subcommand_groups.get(data["data"]["name"])
        if x_base:
            x = x_base.get(data["data"]["options"][0]["name"])

            op = data["data"]["options"][0]
            while op["type"] != 1:
                op = op["options"][0]
            fixed_options = op.get("options", [])
            
            if x is None:
                x = x_base.get(op["name"])
                
            options = await handle_options(data, fixed_options, self.self.parse_method, self._discord)
            if x:
                await x.callback(SlashedSubCommand(self._discord, x, data, user, channel, x.guild_ids), **options)
                return

    async def add_commands(self):
        commands = {
            "globals": {},
            "guilds": {}
        }
        own_guild_ids = [str(x.id) for x in self._discord.guilds]
        
        #region contgext commands
        for command in list(self.context_commands["message"].values()) + list(self.context_commands["user"].values()):
            if command.guild_ids is not MISSING:        
                for x in (command.guild_ids or command.guild_permissions.keys()):
                    if str(x) not in own_guild_ids:
                        raise InvalidArgument("Client is not in a server with the id '" + str(x) + "'")
                
                    if commands["guilds"].get(x) is None:
                        commands["guilds"][x] = {}

                    if command.guild_permissions is not MISSING:
                        command.permissions = command.guild_permissions.get(x)
                    await self.add_guild_command(command, x)
                    commands["guilds"][x][command.name] = command
            else:
                await self.add_global_command(command)
                commands["globals"][command.name] = command
        #endregion

        for command in list(self.commands.values()):
            if command.guild_ids is not MISSING:        
                for x in (command.guild_ids or command.guild_permissions.keys()):
                    if str(x) not in own_guild_ids:
                        raise InvalidArgument("Client is not in a server with the id '" + str(x) + "'")
                
                    if commands["guilds"].get(x) is None:
                        commands["guilds"][x] = {}

                    if command.guild_permissions is not MISSING:
                        command.permissions = command.guild_permissions.get(x)
                    await self.add_guild_command(command, x)
                    commands["guilds"][x][command.name] = command
            else:
                await self.add_global_command(command)
                commands["globals"][command.name] = command

        for base_name in self.subcommands:
            subcommands = list(self.subcommands[base_name].values())
            base = SlashCommand(None, subcommands[0].base_name, options=[x.to_dict() for x in subcommands], guild_ids=subcommands[0].guild_ids, default_permission=subcommands[0].default_permission, guild_permissions=subcommands[0].guild_permissions)
            if base.guild_ids is not MISSING or base.guild_permissions is not MISSING:
                for x in (base.guild_ids or base.guild_permissions.keys()):
                    if str(x) not in own_guild_ids:
                        raise InvalidArgument("Client is not in a server with the id '" + str(x) + "'")

                    if commands["guilds"].get(x) is None:
                        commands["guilds"][x] = {}
                    await self.make_sub_command(base, guild_id=x)
                    commands["guilds"][x][base.name] = base
            else:
                await self.make_sub_command(base)
                commands["globals"][base.name] = base
        
        for base_name in self.subcommand_groups:
            subcommand_groups = list(self.subcommand_groups[base_name].values())
            base = SlashCommand(None, subcommand_groups[0].base_names[0], guild_ids=subcommand_groups[0].guild_ids, options=[], default_permission=subcommand_groups[0].default_permission, guild_permissions=subcommand_groups[0].guild_permissions)
            
            for x in subcommand_groups:
                del x.base_names[0]
                
                subs: List[dict] = []
                for _base in x.base_names:
                    subs.append(SlashOption(OptionType.SUB_COMMAND_GROUP, _base).to_dict())

                subs.append(x.to_dict())

                for i in range(0, len(subs) - 1):
                    subs[i]["options"] = [subs[i + 1]]

                if subs[0]["name"] not in [x["name"] for x in base.options]:
                    base.options += [subs[0]]
                else:
                    b = base.options
                    b[get_index(b, subs[0]["name"], lambda x: x.get('name'))]["options"].append(subs[1])
                    base.options = b

            if base.guild_ids is not MISSING or base.guild_permissions is not MISSING:
                for x in (base.guild_ids or base.guild_permissions.keys()):
                    if str(x) not in own_guild_ids:
                        raise InvalidArgument("Client is not in a server with the id '" + str(x) + "'")
                    
                    if commands["guilds"].get(x) is None:
                        commands["guilds"][x] = {}

                    await self.make_sub_command(base, guild_id=x)
                    commands["guilds"][x][base.name] = base
            else:
                await self.make_sub_command(base)
                commands["globals"][base.name] = base


        if self.delete_unused:
            api_coms = await self._get_global_commands()
            for apic in api_coms:
                if commands["globals"].get(apic["name"]) is None:
                    await delete_global_command(self._discord, apic["id"])
            async for x in self._discord.fetch_guilds():
                _id = str(x.id)
                api_coms = await self._get_guild_commands(_id)
                for apic in api_coms:
                    if commands["guilds"].get(_id) is None or commands["guilds"][_id].get(apic["name"]) is None:
                        await delete_guild_command(self._discord, apic["id"], _id)

        # print("synchronized slash commands")
    
    async def _get_api_command(self, name) -> Union[dict, None]:
        for x in await self._get_commands():
            if x["name"] == name:
                return x
    async def _get_guild_api_command(self, name, guild_id) -> Union[dict, None]:
        for x in await self._get_guild_commands(guild_id):
            if x["name"] == name:
                return x
    async def _get_global_api_command(self, name) -> Union[dict, None]:
        for x in await self._get_global_commands():
            if x["name"] == name:
                return x

    async def _get_commands(self) -> List[dict]:
        return await self._get_global_commands() + await self._get_all_guild_commands()
    async def _get_global_commands(self) -> List[dict]:
        return await get_global_commands(self._discord)
    async def _get_all_guild_commands(self):
        commands = []
        async for x in self._discord.fetch_guilds():
            try:
                commands += await get_guild_commands(self._discord, x.id)
            except Forbidden:
                print("Got forbidden in", x.name, x.id)
                continue
        return commands
    async def _get_guild_commands(self, guild_id: str) -> List[dict]:
        return await get_guild_commands(self._discord, guild_id)
    
    async def add_global_command(self, base):
        """Adds a slash command to the global bot commands
        
        Parameters
        ----------
            base: :class:`~SlashCommand`
                The slash command to add
        
        """
        api_command = await self._get_global_api_command(base.name)
        if api_command is None:
            await create_global_command(base.to_dict(), self._discord)
        else:
            if api_command != base:
                if api_command.get("guild_id") is None:
                    await edit_global_command(api_command["id"], self._discord, base.to_dict())
                else:
                    await delete_guild_command(self._discord, api_command["id"], api_command["guild_id"])
                    await self.add_global_command(base)
    async def add_guild_command(self, base, guild_id):
        """Adds a slash command to a guild
        
        Parameters
        ----------
            base: :class:`~SlashCommand`
                The guild slash command which should be added
            guild_id: :class:`str`
                The ID of the guild where the command is going to be added
        
        """
        target_guild = guild_id
        api_command = await self._get_guild_api_command(base.name, guild_id)
        # If no command in that guild
        if api_command is None:
            # Check global commands
            api_command = await self._get_global_api_command(base.name)
            # If global command exists, it will be deleted
            if api_command is not None:
                await delete_global_command(self._discord, api_command["id"])
            await create_guild_command(base.to_dict(), self._discord, target_guild, base.permissions.to_dict())
        else:
        # elif api_command != base:
            await edit_guild_command(api_command["id"], self._discord, target_guild, base.to_dict(), base.permissions.to_dict())

    async def make_sub_command(self, base: SlashCommand, guild_id=MISSING):
        """Creates a new sub command and edits it if the base already exsits

        Parameters
        ----------
            base: :class:`~SlashCommand`
                The slash command with the sub commands in it
            guild_id: :class:`str` | :class:`int`
                The guild id where the command should be useable
        
        """
        if guild_id is not MISSING:
            await self.add_guild_command(base, guild_id)
        else:
            await self.add_global_command(base)
            

    async def delete_global_commands(self):
        """**Deletes all global commands**"""
        await delete_global_commands(self._discord)
    async def delete_guild_commands(self, guild_id: str):
        """
        **Deletes all commands in a guild**

        Parameters
        ----------
            guild_id: :class:`str`
                The id of the guild where all commands are going to be deleted
        
        """
        await delete_guild_commands(self._discord, guild_id)
    async def nuke_commands(self):
        """**Deletes every command for the bot, including globals and commands in every guild**"""
        print("nuking...")
        await self.delete_global_commands()
        print("nuked global commands")
        async for guild in self._discord.fetch_guilds():
            print("nuking commands in", guild.id)
            await self.delete_guild_commands(guild.id)
            print("nuked commands in", guild.id, guild.name)
        print("nuked")


    def command(self, name, description=MISSING, options=MISSING, guild_ids=MISSING, default_permission=True, guild_permissions=MISSING):
        """A decorator for a slash command
        
        command in discord:
            ``/name [options]``

        Parameters
        ----------
            name: :class:`str`
                1-32 characters long name
                
                .. note::

                    The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
            
            description: :class:`str`, optional
                1-100 character description of the command; default the command name
            options: List[:class:`~SlashOptions`], optional
                The parameters for the command; default MISSING
            choices: :class:`[type]`, optional
                Choices for string and int types for the user to pick from; default MISSING
            guild_ids: List[:class:`str` | :class:`int`], optional
                A list of guild ids where the command is available; default MISSING
            default_permission: :class:`bool`, optional
                Whether the command can be used by everyone or not
            guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
                The permissions for the command in guilds
                    Format: ``{"guild_id": SlashPermission}``

        Decorator
        ---------

            callback: :class:`method(ctx)`
                The asynchron function that will be called if the command was used
                    ctx: :class:`~SlashedCommand`
                        The used slash command

                    .. note::

                        ``ctx`` is just an example name, you can use whatever you want for that

        Example
        -------
        .. code-block::

            @slash.command(name="hello_world", description="This is a test command", 
            options=[
                SlashOption(str, name="parameter", description="this is a parameter", choices=[{ "name": "choice 1", "value": "test" }])
            ], guild_ids=["785567635802816595"], default_permission=False, 
            guild_permissions={
                    "785567635802816595": SlashPermission(allowed={"539459006847254542": SlashPermission.USER})
                }
            )
            async def command(ctx, parameter = None):
                ...
        """
        def wrapper(callback):
            """The wrapper for the callback function. The function's parameters have to have the same name as the parameters specified in the slash command.

            `ctx` is of type :class:`~SlashedCommand` and is used for responding to the interaction and more

            Examples
            --------
            - no parameter:
                `async def command(ctx): ...`
            - required parameter "number":
                `async def command(ctx, number): ...`
            - optional parameter "user":
                `async def command(ctx, user=default_value)`
            - multiple optional parameters "user", "number":
                `async def command(ctx, user=default_value, number=default_value)`
            - one required and one optional parameter "user", "text":
                `async def command(ctx, user, text=default_value)`

            Note: Replace `default_value` with a value you want to be used if the parameter is not specified in discord, if you don't want a default value, just set it to `None`
            """
            self.commands[name] = SlashCommand(callback, name, description, options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
        return wrapper
    def subcommand(self, base_name, name, description=MISSING, options=MISSING, guild_ids=MISSING, default_permission=True, guild_permissions=MISSING):
        """A decotator for a subcommand
        
        command in discord
            ``/base_name name [options]``

        Parameters
        ----------
            base_names: List[:class:`str`]
                The names of the parent base
            name: :class:`str`
                1-32 characters long name
                .. note::

                    The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
            description: :class:`str`, optional
                1-100 character description of the command; default the command name
            options: List[:class:`~SlashOptions`], optional
                The parameters for the command; default MISSING
            guild_ids: List[:class:`str` | :class:`int`], optional
                A list of guild ids where the command is available; default MISSING
            default_permissions: :class:`bool`, optional
                Whether the command can be used by everyone or not
            guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
                The permissions for the command in guilds
                    Format: ``{"guild_id": SlashPermission}``

            .. note::

                Permissions will be the same for every subcommand with the same base 

        Decorator
        ---------

            callback: :class:`method(ctx)`
                The asynchron function that will be called if the command was used
                    ctx: :class:`~SlashedSubCommand`
                        The used slash command

                    .. note::

                        ``ctx`` is just an example name, you can use whatever you want for that

        Example
        -------
        .. code-block::

            @slash.subcommand(base_name="hello", name="world", options=[
                SlashOption(argument_type="user", name="user", description="the user to tell the holy words")
            ], guild_ids=["785567635802816595"])
            async def command(ctx, user):
                ...
        """
        def wrapper(callback):
            """The wrapper for the callback function. The function's parameters have to have the same name as the parameters specified in the slash command.

            `ctx` is of type :class:`~SlashedCommand` and is used for responding to the interaction and more

            Examples
            --------
            - no parameter:
                `async def command(ctx): ...`
            - required parameter "number":
                `async def command(ctx, number): ...`
            - optional parameter "user":
                `async def command(ctx, user=default_value)`
            - multiple optional parameters "user", "number":
                `async def command(ctx, user=default_value, number=default_value)`
            - one required and one optional parameter "user", "text":
                `async def command(ctx, user, text=default_value)`

            Note: Replace `default_value` with a value you want to be used if the parameter is not specified in discord, if you don't want a default value, just set it to `None`
            """
            if self.subcommands.get(base_name) is None:
                self.subcommands[base_name] = {}

            self.subcommands[base_name][name] = SubSlashCommand(callback, base_name, name, description, options=options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
        return wrapper
    def subcommand_group(self, base_names, name, description=MISSING, options=MISSING, guild_ids=MISSING, default_permission=True, guild_permissions=MISSING):
        """A decorator for a subcommand group
        
        command in discord
            ``/base_names... name [options]``

        Parameters
        ----------
            base_names: List[:class:`str`]
                The names of the parent bases, currently limited to 2
            name: :class:`str`
                1-32 characters long name
                .. note::

                    The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
            description: :class:`str`, optional
                1-100 character description of the command; default the command name
            options: List[:class:`~SlashOptions`], optional
                The parameters for the command; default MISSING
            choices: :class:`[type]`, optional
                Choices for string and int types for the user to pick from; default MISSING
            guild_ids: List[:class:`str` | :class:`int`], optional
                A list of guild ids where the command is available; default MISSING
            default_permission: :class:`bool`, optional
                Whether the command can be used by everyone or not
            guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
                The permissions for the command in guilds
                    Format: ``{"guild_id": SlashPermission}``

            .. note::

                Permissions will be the same for every subcommand with the same base

        Decorator
        ---------

            callback: :class:`method(ctx)`
                The asynchron function that will be called if the command was used
                    ctx: :class:`~SlashedSubCommand`
                        The used slash command

                    .. note::

                        ``ctx`` is just an example name, you can use whatever you want for that
        
        Example
        -------
        .. code-block::

            @slash.subcommand_group(base_names=["hello", "beautiful"], name="world", options=[
                SlashOption(argument_type="user", name="user", description="the user to tell the holy words")
            ], guild_ids=["785567635802816595"])
            async def command(ctx, user):
                ...

        """
        def wrapper(callback):
            """The wrapper for the callback function. The function's parameters have to have the same name as the parameters specified in the slash command.

            `ctx` is of type :class:`~SlashedCommand` and is used for responding to the interaction and more

            Examples
            --------
            - no parameter:
                `async def command(ctx): ...`
            - required parameter "number":
                `async def command(ctx, number): ...`
            - optional parameter "user":
                `async def command(ctx, user=default_value)`
            - multiple optional parameters "user", "number":
                `async def command(ctx, user=default_value, number=default_value)`
            - one required and one optional parameter "user", "text":
                `async def command(ctx, user, text=default_value)`

            Note: Replace `default_value` with a value you want to be used if the parameter is not specified in discord, if you don't want a default value, just set it to `None`
            """
            if self.subcommand_groups.get(base_names[0]) is None:
                self.subcommand_groups[base_names[0]] = {}
            self.subcommand_groups[base_names[0]][name] = SubSlashCommandGroup(callback, base_names, name, description, options=options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)

        return wrapper
    def user_command(self, name, guild_ids, default_permission=True, guild_permissions = MISSING):
        """Decorator for user context commands in discord.
            ``Right-click username`` -> ``apps`` -> ``commands is displayed here``


        Parameters
        ----------
            name: :class:`str`
                The name of the command
            guild_ids: List[:class:`str` | :class:`int`]
                A list of guilds where the command can be used
            default_permission: :class:`bool`, optional
                Whether the command can be used by everyone; default True
            guild_permissions: Dict[:class:`SlashPermission`], optional
                Special permissions for guilds; default MISSING

        Decorator
        ---------

            callback: :class:`method(ctx, user)`
                The asynchron function that will be called if the command was used
                    ctx: :class:`~SlashedSubCommand`
                        The used slash command
                    user: :class:`discord.Member`
                        The user on which the command was used
                    .. note::

                        ``ctx`` and ``user`` are just example names, you can use whatever you want for that
        
        Example
        -------
        
        .. code-block::

            @slash.user_command(name="call", guild_ids=["785567635802816595"], default_permission=False, guild_permissions={
                "785567635802816595": SlashPermission(allowed={
                    "585567635802816595": SlashPermission.USER
                })
            })
            async def call(ctx, message):
                ...
        """
        def wraper(callback):
            self.context_commands["user"][name] = UserCommand(callback, name, guild_ids, default_permission, guild_permissions)
        return wraper
    def message_command(self, name, guild_ids, default_permission=True, guild_permissions = MISSING):
        """Decorator for message context commands in discord.
            ``Right-click message`` -> ``apps`` -> ``commands is displayed here``


        Parameters
        ----------
            name: :class:`str`
                The name of the command
            guild_ids: List[:class:`str` | :class:`int`]
                A list of guilds where the command can be used
            default_permission: :class:`bool`, optional
                Whether the command can be used by everyone; default True
            guild_permissions: Dict[:class:`SlashPermission`], optional
                Special permissions for guilds; default MISSING

        Decorator
        ---------

            callback: :class:`method(ctx, message)`
                The asynchron function that will be called if the command was used
                    ctx: :class:`~SlashedSubCommand`
                        The used slash command
                    message: :class:`~Message`
                        The message on which the command was used
                    .. note::

                        ``ctx`` and ``message`` are just example names, you can use whatever you want for that
        
        Example
        -------
        
        .. code-block::

            @slash.message_command(name="quote", guild_ids=["785567635802816595"], default_permission=False, guild_permissions={
                "785567635802816595": SlashPermission(allowed={
                    "585567635802816595": SlashPermission.USER
                })
            })
            async def quote(ctx, message):
                ...
        """
        def wraper(callback):
            self.context_commands["message"][name] = MessageCommand(callback, name, guild_ids, default_permission, guild_permissions)
        return wraper

class Components():
    """A class for using and receiving message components in discord
    
    Parameters
    -----------
        client: :class:`discord.Client`
            The main discord client
        receive_events: :class:`bool`
            Whether events like components should be received

    Example
    ------------------
    Example for using the listener

    
    .. code-block::

        ...
        # Your bot declaration should be here
        components = components(client)
        
    
    for listening to button presses, use
    
    .. code-block::

        ...
        @client.event("on_button_press")
        async def on_button(pressedButton, message):
            ...

    for listening to select menu selections, use
    

    .. code-block::

        ...
        @client.event("on_menu_select")
        async def on_select(seletedMenu, message):
            ...
        
    For components that will listen to a custom id, use

    .. code-block::

        ...
        @components.listening_component(custom_id="custom_id_here")
        async def my_func(component, msg):
            ...
    """
    def __init__(self, client: com.Bot, receive_events = True):
        """Creates a new compnent listener
        
        Example
        ```py
        Components(client)
        ```
        """
        self._listening_components = []
        """A list of components that are listening for interaction"""
        self._discord = client
        if receive_events is True:
            self._discord.add_listener(self._on_socket_response, 'on_socket_response')
    
    async def _on_socket_response(self, msg):
        """Will be executed if the bot receives a socket response"""
        if msg["t"] != "INTERACTION_CREATE":
            return
        data = msg["d"]

        if data["type"] != 3:
            return

        guild = await self._discord.fetch_guild(data["guild_id"])
        user = discord.Member(data=data["member"], guild=guild, state=self._discord._connection)
        

        msg = await getResponseMessage(self._discord._connection, data=data, application_id=self._discord.user.id, user=user, response=True)
        component = EphemeralComponent(self._discord.user.id, state=self._discord._connection, user=user, data=data) if data["message"]["flags"] == 64 else msg.interaction_component

        # Get listening components with the same custom id
        _listening_components = [x for x in self._listening_components if data["data"]["custom_id"] == x[1]]
        if len(_listening_components) == 1:
            await _listening_components[0][0](component, msg)

        if data["data"]["component_type"] == 2:
            self._discord.dispatch("button_press", component, msg)
        elif data["data"]["component_type"] == 3:
            self._discord.dispatch("menu_select", component, msg)

    async def send(self, channel, content=MISSING, *, tts=False, embed=MISSING, embeds=MISSING, file=MISSING, 
            files=MISSING, delete_after=MISSING, nonce=MISSING, allowed_mentions=MISSING, reference=MISSING, 
            mention_author=MISSING, components=MISSING) -> Message:
        """Sends a message to a textchannel

        Parameters
        ----------
        channel: :class:`discord.TextChannel` | :class:`int` | :class:`str`
            The target textchannel or the id of it
        content: :class:`str`, optional
            The message text content; default None
        tts: :class:`bool`, optional
            True if this is a text-to-speech message; default False
        embed: :class:`discord.Message`, optional
            Embedded rich content (up to 6000 characters)
        embeds: List[:class:`discord.Embed`], optional
            Up to 10 embeds; default None
        file: :class:`discord.File`, optional
            A file sent as an attachment to the message; default None
        files: List[:class:`discord.File`], optional
            A list of file attachments; default None
        delete_after: :class:`float`, optional
            After how many seconds the message should be deleted; default None
        nonce: :class:`int`, optional
            The nonce to use for sending this message. If the message was successfully sent, then the message will have a nonce with this value; default None
        allowed_mentions: :class:`discord.AllowedMentions`, optional
            A list of mentions proceeded in the message; default None
        reference: :class:`discord.MessageReference` | :class:`discord.Message`, optional
            A message to refer to (reply); default None
        mention_author: :class:`bool`, optional
            True if the author should be mentioned; default None
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`], optional
            A list of message components included in this message; default None

        Returns
        -------
        :return: Returns the sent message
        :type: :class:`~Message`

        Raises
        ------
        :raises: :class:`discord.InvalidArgument`: A passed argument was invalid
        """

        if type(channel) not in [discord.TextChannel, int, str]:
            raise discord.InvalidArgument("Channel must be of type discord.TextChannel")

        channel_id = channel.id if type(channel) is discord.TextChannel else channel
        payload = jsonifyMessage(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, reference=reference, mention_author=mention_author, components=components)

        route = BetterRoute("POST", f"/channels/{channel_id}/messages")

        r = None
        if file is MISSING and files is MISSING:
            r = await self._discord.http.request(route, json=payload)
        else:
            r = await send_files(route, files=files or [file], payload=payload, http=self._discord.http)

        msg = Message(state=self._discord._get_state(), channel=channel, data=r)
            
        if delete_after is not None:
            await msg.delete(delay=delete_after)
        
        return msg
    def send_webhook(self, webhook, content=MISSING, *, wait=False, username=MISSING, avatar_url=MISSING, tts=False, files=MISSING, embed=MISSING, embeds=MISSING, allowed_mentions=MISSING, components=MISSING) -> Union[WebhookMessage, None]:
        """Sends a webhook message
        
        Parameters
        ----------
            webhook: :class:`discord.Webhook`
                The webhook which will send the message
            content: :class:`str`, optional
                the message contents (up to 2000 characters); default None
            wait: :class:`bool`, optional
                if `True`, waits for server confirmation of message send before response, and returns the created message body; default False
            username: :class:`str`, optional
                override the default username of the webhook; default None
            avatar_url: :class:`str`, optional
                override the default avatar of the webhook; default None
            tts: :class:`bool`, optional
                true if this is a TTS message; default False
            files: :class:`discord.File`
                A list of files which will be sent as attachment
            embed: :class:`discord.Embed`
                Embed rich content, optional
            embeds: List[:class:`discord.Embed`], optional
                embedded rich content; default None
            allowed_mentions: :class:`discord.AllowedMentions`, optional
                allowed mentions for the message; default None
            components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`], optional
                the message components to include with the message; default None
        Returns
        -------
            :returns: The message sent, if wait was True, else nothing will be returned
            :type: :class:`~WebhookMessage` | :class:`None`
        
        """
        payload = jsonifyMessage(content, tts=tts, embed=embed, embeds=embeds, allowed_mentions=allowed_mentions, components=components)
        payload["wait"] = wait
        if username is not None:
            payload["username"] = username
        if avatar_url is not None:
            payload["avatar_url"] = str(avatar_url)

        return webhook._adapter.execute_webhook(payload=payload, wait=wait, files=files)
    def listening_component(self, custom_id):
        """A decorator for a listening component, that will be always executed if the invoked interaction has the custom_id passed

        .. Warning::

            The button_press event will not be dispatched when using ``listening_component``, so make sure you don't use that custom_id here for any of your normal components
        
        Parameters
        ----------
            custom_id: :class:`str`
                The custom_id of the components to listen to

        Decorator
        ---------
            callback: :class:`method(component, message)`
                The asynchron function that will be called if a component with the custom_id was invoked

                There will be two parameters passed

                    component: :class:`~PressedButton` or :class:`~SelectedMenu`
                        The invoked component
                    message: :class:`~ResponseMessage`
                        The message in which the component was invoked

                    .. note::

                        ``component`` and ``message`` are just example names, you can use whatever you want for them

        
        """
        def wrapper(callback):
            if len([x for x in self._listening_components if x[1] == custom_id]) > 0:
                raise Exception("custom_id " + str(custom_id) + " is already in use! Use another custom_id")
            self._listening_components.append((callback, custom_id))
        return wrapper
    

class UI():
    """The main extension for the package to use slash commands and message components
        
        Parameters
        ----------
            client: :class:`discord.ext.commands.Bot`
                The discord bot client

            slash_options: :class:`dict`, optional
                Settings for the slash command part; Default `{parse_method: ParseMethod.AUTO, delete_unused: False, wait_sync: 1}`
                
                ``parse_method``: :class:`int`, optional
                    How the received interaction argument data should be treated; Default ``ParseMethod.AUTO``


                ``delete_unused``: :class:`bool`, optional
                    Whether the commands that are not registered by this slash ui should be deleted in the api; Default ``False``
        
                ``wait_sync``: :class:`float`, optional
                    How many seconds will be waited until the commands are going to be synchronized; Default ``1``


    """
    def __init__(self, client, slash_options = {"parse_method": ParseMethod.AUTO, "delete_unused": False, "wait_sync": 1}) -> None:
        """Creates a new ui object
        
        Example
        ```py
        UI(client, slash_options={"delete_unused": True, "wait_sync": 2})
        ```
        """
        self.components = Components(client)
        """For using message components
        
        :type: :class:`~Components`
        """
        if slash_options is None:
            slash_options = {"resolve_data": True, "delete_unused": False, "wait_sync": 1}
        self.slash = Slash(client, **slash_options)
        """For using slash commands
        
        :type: :class:`~Slash`
        """



import typing
from ..tools import MISSING

import inspect
import discord
from discord.errors import InvalidArgument

class SlashOption():
    """An option for a slash command
        
        Parameters
        ----------
            argument_type: :class:`int` | :class:`str` | :class:`class`
                What type of parameter the option should accept
            name: :class:`str`
                1-32 lowercase character name for the option
            description: :class:`str`, optional
                1-100 character description of the command; default name
            required: :class:`bool`, optional
                If the parameter is required or optional; default False
            choices: List[:class:`dict`], optional
                Choices for string and int types for the user to pick from; default MISSING
                    Choices should be formated like this: ``[{"name": "name of the choice", "value": "the real value"}, ...]``

                    .. note::

                        The choice value has to be of the same type as the type this option accepts

            options: :class:`~SlashOption`
                This parameter is only for subcommands to work, you shouldn't need to use that, unless you know what you're doing 
        """
    def __init__(self, argument_type, name, description=MISSING, required=False, choices=MISSING, options=MISSING) -> None:
        """
        Creates a new option for a slash command

        Example
        ```py
        SlashOption(argument_type=int, name="Your number", required=True, choices=[{"name": "a cool number", "value": 2}])
        ```
        """
        self._json = {}
        self.argument_type = argument_type
        self.name = name
        self.description = description or name
        if required is True:
            self.required = required
        
        if options is not MISSING:
            self.options = options
        if choices is not MISSING:
            self._json["choices"] = choices

    @property
    def argument_type(self) -> int:
        """Parameter type that the option accepts
        
        :type: :class:`int`
        """
        return self._json["type"]
    @argument_type.setter
    def argument_type(self, value):
        self._json["type"] = OptionType.any_to_type(value)

    @property
    def name(self) -> str:
        """The name of the option appearing in discord
        
        :type: :class:`str`
        """
        return self._json["name"]
    @name.setter
    def name(self, value):
        if len(value) > 32 or len(value) < 1:
            raise InvalidArgument("name must be between 1 and 32 characters")
        self._json["name"] = value.lower().replace(" ", "_")

    @property
    def description(self) -> str:
        """The description of the option appearing under the name
        
        :type: :class:`str`
        """
        return self._json["description"]
    @description.setter
    def description(self, value):
        if len(value) > 100 or len(value) < 1:
            raise InvalidArgument("description must be between 1 and 100 characters")
        self._json["description"] = value

    @property
    def required(self) -> bool:
        """Whether this parameter is required to use or not

        :type: :class:`bool`
        """
        return self._json.get("required", False)
    @required.setter
    def required(self, value):
        if type(value) is not bool:
            raise InvalidArgument("required has to be of type bool, not " + str(type(value))) 
        self._json["required"] = value

    @property
    def choices(self) -> typing.List[dict]:
        """Choices for string and int types for the user to pick from
        
        .. note::
    
            Choices are formated like this: ``[{"name": "name of the choice", "value": "the real value"}, ...]`` 
    
        :type: List[:class:`dict`]
        """
        return self._json.get("choices")
    @choices.setter
    def choices(self, value):
        self._json["choices"] = value

    @property
    def options(self) -> typing.List['SlashOption']:
        return [SlashOption(x["type"], x["name"], x["description"], x.get("required", False), x.get("choices"), x.get("options")) for x in self._json.get("options")]
    @options.setter
    def options(self, options):
        if type(options) is list and all(type(x) is SlashOption for x in options):
            self._json["options"] = [x.to_dict() for x in options]
        elif type(options) is list and all(type(x) is dict for x in options):
            self._json["options"] = options
        else:
            raise InvalidArgument("'options' has to be of type List[dict] or List[dict], not " + str(type(options)))


    def to_dict(self):
        return self._json

class OptionType:
    """The list of possible slash command option types"""

    SUB_COMMAND             =          1
    SUB_COMMAND_GROUP       =          2
    STRING                  =          3
    INTEGER                 =          4
    BOOLEAN                 =          5
    MEMBER                  =          6
    CHANNEL                 =          7
    ROLE                    =          8
    MENTIONABLE             =          9
    FLOAT                   =         10

    @classmethod
    def any_to_type(cls, whatever):
        if type(whatever) is int and whatever in range(1, 11):
            return whatever
        if inspect.isclass(whatever):
            if whatever is str:
                return cls.STRING
            if whatever is int:
                return cls.INTEGER
            if whatever is bool:
                return cls.BOOLEAN
            if whatever in [discord.User, discord.Member]:
                return cls.USER
            if whatever is discord.TextChannel:
                return cls.CHANNEL
            if whatever is discord.Role:
                return cls.ROLE
            if whatever is float:
                return cls.FLOAT
        if type(whatever) is str:
            whatever = whatever.lower()
            if whatever in ["str", "string"]:
                return cls.STRING
            if whatever in ["int", "integer"]:
                return cls.INTEGER
            if whatever in ["bool", "boolean"]:
                return cls.BOOLEAN
            if whatever in ["user", "discord.user", "member", "discord.member", "usr", "mbr"]:
                return cls.USER
            if whatever in ["channel", "textchannel", "discord.textchannel", "txtchannel"]:
                return cls.CHANNEL
            if whatever in ["role", "discord.role"]:
                return cls.ROLE
            if whatever in ["mentionable", "mention"]:
                return cls.MENTIONABLE
            if whatever in ["float", "floating", "floating number", "f"]:
                return cls.FLOAT

        
class SlashPermission():
    """Permissions for a slash commannd
        
        Parameters
        ----------
            allowed: :class:`dict`, optional
                A list of ids that can use the command; default MISSING
                    Format: ``{"role_or_user_id": permission_type}]``


            forbidden: :class:`dict`, optional
                A list of ids that are forbidden to use the command; default MISSING
        

            .. note::

                If you want to use a role id, the permission type has to be 1, and if you want to specify a user id, it has to be 2

                You can use ``SlashPermission.ROLE`` and ``SlashPermission.USER`` instead
        """
    def __init__(self, allowed: dict=MISSING, forbidden=MISSING) -> None:
        """Creates a new permission object for a slash command
        
        Example
        ```py
        SlashPermission(forbidden={
                "785567792899948577": SlashPermission.ROLE,
                "355333222596476930": SlashPermission.USER,
                "539459006847254542": SlashPermission.USER
            }, allowed={
                "539459006847255232": SlashPermission.User
            }
        )
        ```
        """
        
        self._json = []
        if allowed is not MISSING:
            for _id, _type in allowed.items():
                self._json.append(
                    {
                        "id": _id,
                        "type": _type,
                        "permission": True
                    }
                )
        if forbidden is not MISSING:
            for _id, _type in forbidden.items():
                self._json.append(
                    {
                        "id": _id,
                        "type": _type,
                        "permission": False
                    }
                )

    def to_dict(self):
        return self._json
    
    ROLE = 1
    USER = 2

    @property
    def allowed(self) -> typing.List[typing.Union[str, int]]:
        return [x["id"] for x in self._json if x["permission"] == True]
    @property
    def forbidden(self) -> typing.List[typing.Union[str, int]]:
        return [x["id"] for x in self._json if x["permission"] == False]


class SlashCommand():
    """A basic slash command
        
        Parameters
        ----------
            name: :class:`str`
                1-32 characters long name
                .. note::

                    The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
            description: :class:`str`, optional
                1-100 character description of the command; default the command name
            options: List[:class:`~SlashOptions`], optional
                Parameters for the command; default MISSING
            choices: :class:`[type]`, optional
                Choices for string and int types for the user to pick from; default MISSING
            guild_ids: :class:`str` | :class:`int`, optional
                A list of guild ids where the command is available; default MISSING
            default_permission: :class:`bool`
                Whether the command should be usable for everyone or not
            guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
                The permissions for the command in guilds
                    Format: ``{"guild_id": SlashPermission}``


    """
    def __init__(self, callback, name, description=MISSING, options=MISSING, guild_ids=MISSING, default_permission=MISSING, guild_permissions=MISSING) -> None:
        """
        Creates a new base slash command
        
        Example

        ```py
        async def my_function(command, parameter=None):
            pass

        SlashCommand(callback=my_function, name="hello_world", description="This is a test command", 
            options=[
                SlashOption(str, name="parameter", description="this is a parameter", choices=[{ "name": "choice 1", "value": 1 }])
            ], guild_ids=["785567635802816595"], default_permission=False, 
            guild_permissions={ 
                "785567635802816595": SlashPermission(allowed={"539459006847254542": SlashPermission.USER}) 
            })
        ```
        """
        self._json = {
            "type": 1
        }
        self.name = name
        self.description = description or name
        if callback is not None:
            if not inspect.iscoroutinefunction(callback):
                raise InvalidArgument("callback has to be async")

            callback_params = inspect.signature(callback).parameters
            if options is not MISSING:
                for op in options:
                    if callback_params.get(op.name) is None:
                        raise InvalidArgument("Missing parameter '" + op.name + "' in callback function")
                    param = callback_params[op.name]
                    if not op.required and param.default is param.empty:
                        raise InvalidArgument("Parameter '" + param.name + "' in callback function needs to be optional (" + param.name + "=None)")

        self.callback = callback

        if options is not MISSING:
            self.options = options
        if default_permission is not MISSING:
            self.default_permission = default_permission
        self.guild_permissions: typing.Dict[(str, SlashPermission)] = guild_permissions or MISSING
        self.permissions = SlashPermission()
        
        self.guild_ids = guild_ids

    # region command
    @property
    def name(self) -> str:
        """The name of the slash command
        
        :type: :class:`str`
        """
        return self._json["name"]
    @name.setter
    def name(self, value):
        if len(value) > 32 or len(value) < 1:
            raise InvalidArgument("name must be between 1 and 32 characters")
        self._json["name"] = str(value).lower().replace(" ", "-")
    @property
    def description(self) -> str:
        """The description of the command
        
        :type: :class:`str`
        """
        return self._json['description']
    @description.setter
    def description(self, value):
        if len(value) > 100 or len(value) < 1:
            raise InvalidArgument("description must be between 1 and 100 characters")
        self._json["description"] = value
    @property
    def options(self) -> typing.List['SlashOption']:
        """The parameters for the command
        
        :type: List[:class:`~SlashOption`]
        """
        return self._json.get("options")
    @options.setter
    def options(self, options):
        if not type(options) is list:
            raise TypeError("options must be of type List[dict] or List[SlashOptions], not " + str(type(options)))
        if all(type(x) is SlashOption for x in options):
            self._json["options"] = [x.to_dict() for x in options]
        elif all(type(x) is dict for x in options):
            self._json["options"] = options
        else:
            raise InvalidArgument("options must be of type List[dict] or List[SlashOptions], not " + str(type(options)))
    # endregion
    # region permissions
    @property
    def default_permission(self):
        return self._json["default_permission"]
    @default_permission.setter
    def default_permission(self, value):
        self._json["default_permission"] = value
    # endregion

    def to_dict(self):
        return self._json
    
    def __eq__(self, o: object) -> bool:
        if type(o) is dict:
            return (
                o.get('name') == self.name
                and o.get('description') == self.description
                and o.get('options') == self.options
            )
        elif type(o) is SlashCommand:
            return (
                o.name == self.name
                and o.description == self.description
                and o.options == self.options
            )
        else:
            return False


class ContextCommand():
    def __init__(self) -> None:
        self._json = {}

    @property
    def name(self):
        return self._json["name"]
    @name.setter
    def name(self, value):
        self._json["name"] = value
    @property
    def default_permission(self):
        return self._json["default_permission"]
    @default_permission.setter
    def default_permission(self, value):
        self._json["default_permission"] = value

    def to_dict(self):
        return self._json

class UserCommand(ContextCommand):
    def __init__(self, callback, name, guild_ids = MISSING, default_permission = True, guild_permissions = MISSING) -> None:
        super().__init__()
        self._json = {
            "type": 2
        }
        if callback is not None:
            if not inspect.iscoroutinefunction(callback):
                raise InvalidArgument("callback has to be async")

            callback_params = inspect.signature(callback).parameters
            if not len(callback_params) >= 2:
                raise InvalidArgument("Callback function has to take at least 2 parameters")

        self.callback = callback
        self.name = name
        self.guild_ids = guild_ids
        self.default_permission = default_permission
        self.guild_permissions = guild_permissions
        self.permissions = SlashPermission()

class MessageCommand(ContextCommand):
    def __init__(self, callback, name, guild_ids = MISSING, default_permission = True, guild_permissions = MISSING) -> None:
        super().__init__()
        self._json = {
            "type": 3
        }
        if callback is not None:
            if not inspect.iscoroutinefunction(callback):
                raise InvalidArgument("callback has to be async")

            callback_params = inspect.signature(callback).parameters
            if not len(callback_params) >= 2:
                raise InvalidArgument("Callback function has to take at least 2 parameters")

        self.callback = callback
        self.name = name
        self.guild_ids = guild_ids
        self.default_permission = default_permission
        self.guild_permissions = guild_permissions
        self.permissions = SlashPermission()

class SubSlashCommand(SlashCommand):
    def __init__(self, callback, base_name, name, description=MISSING, options=MISSING, guild_ids=MISSING, default_permission=MISSING, guild_permissions=MISSING) -> None:
        SlashCommand.__init__(self, callback, name, description, options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
        self.base_name = base_name.replace(" ", "-").lower()


    def to_dict(self):
        return SlashOption(OptionType.SUB_COMMAND, self.name, self.description, options=self.options).to_dict()

class SubSlashCommandGroup(SlashCommand):
    def __init__(self, callback, base_names, name, description=MISSING, options=MISSING, guild_ids=MISSING, default_permission=MISSING, guild_permissions=MISSING) -> None:
        if len(base_names) > 2:
            raise InvalidArgument("subcommand groups are currently limited to 2 bases")
        if any([len(x) > 32 or len(x) < 1 for x in base_names]):
            raise InvalidArgument("base_names needs to be between 32 and 1 characters long")
        SlashCommand.__init__(self, callback, name, description, options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
        self.base_names = [x.replace(" ", "-").lower() for x in base_names]
        
    def to_dict(self):
        return SlashOption(OptionType.SUB_COMMAND, self.name, self.description, options=self.options).to_dict()



import aiohttp
from ..tools import MISSING, get
from ..http import BetterRoute

from discord.ext import commands as com
from discord.errors import Forbidden, HTTPException

from requests import put

async def get_command(command_name, client: com.Bot, guild_id=MISSING):
    return get(
        (await get_global_commands(client)) if guild_id is MISSING else (await get_guild_commands(client, guild_id)),
        command_name, lambda x: x.get("name")
    )
async def get_id(command_name, client: com.bot, guild_id=MISSING):
    found = (await get_command(command_name, client, guild_id))
    if found is None:
        raise Exception("No slash command found with name '" + command_name + "'")
    return found.get('id')

async def delete_global_commands(client: com.Bot):
    commands = await client.http.request(BetterRoute("GET", f"/applications/{client.user.id}/commands"))
    for x in commands:
        await delete_global_command(client, x["id"])
       
async def delete_guild_commands(client: com.Bot, guild_id):
    try:
        commands = await client.http.request(BetterRoute("GET", f"/applications/{client.user.id}/guilds/{guild_id}/commands"))
        for x in commands:
            await delete_guild_command(client, x["id"], guild_id)
    except Forbidden:
        print("got forbidden in", guild_id)

async def delete_global_command(client: com.Bot, command_id):
    return await client.http.request(BetterRoute("DELETE", f"/applications/{client.user.id}/commands/{command_id}"))
async def delete_guild_command(client: com.Bot, command_id, guild_id):
    return await client.http.request(BetterRoute("DELETE", f"/applications/{client.user.id}/guilds/{guild_id}/commands/{command_id}"))

async def update_command_permission(id, token, guild_id, command_id, permissions):
    async with aiohttp.ClientSession() as client:
        async with client.put(f"https://discord.com/api/v9/applications/{id}/guilds/{guild_id}/commands/{command_id}/permissions", 
            headers={"Authorization": "Bot " + token}, json={"permissions": permissions}) as response:
            if response.status == 200:
                return await response.json()
            raise HTTPException(response, response.content)
        
async def create_global_command(command: dict, client: com.Bot):
    return await client.http.request(BetterRoute("POST", f"/applications/{client.user.id}/commands"), json=command)
async def create_guild_command(command, client: com.Bot, guild_id, permissions = []):
    data = await client.http.request(BetterRoute("POST", f"/applications/{client.user.id}/guilds/{guild_id}/commands"), json=command)
    return await update_command_permission(client.user.id, client.http.token, guild_id, data["id"], permissions)


async def edit_global_command(command_id: str, client: com.Bot, new_command: dict):
    return await client.http.request(BetterRoute("PATCH", f"/applications/{client.user.id}/commands/{command_id}"), json=new_command)
async def edit_guild_command(command_id, client: com.Bot, guild_id: str, new_command: dict, permissions: dict):
    data = await client.http.request(BetterRoute("PATCH", f"/applications/{client.user.id}/guilds/{guild_id}/commands/{command_id}"), json=new_command)
    return await update_command_permission(client.user.id, client.http.token, guild_id, data["id"], permissions)

async def get_global_commands(client):
    return await client.http.request(BetterRoute("GET", f"/applications/{client.user.id}/commands"))
async def get_guild_commands(client, guild_id):
    return await client.http.request(BetterRoute("GET", f"/applications/{client.user.id}/guilds/{guild_id}/commands"))



from .tools import MISSING
from .client import Components, Slash
import discord
from discord.ext import commands

from .receive import Message, WebhookMessage
from .http import jsonifyMessage, BetterRoute, send_files

import sys
module = sys.modules["discord"]


#region message override
async def send(self: discord.TextChannel, content=None, **kwargs) -> Message:
        payload = jsonifyMessage(content=content, **kwargs)

        channel_id = self.id if type(self) is not commands.Context else self.channel.id
        route = BetterRoute("POST", f"/channels/{channel_id}/messages")
        
        r = None
        if kwargs.get("file") is None and kwargs.get("files") is None:
            r = await self._state.http.request(route, json=payload)
        else:
            r = await send_files(route, files=kwargs.get("files") or [kwargs.get("file")], payload=payload, http=self._state.http)
        
        msg = Message(state=self._state, channel=self if type(self) is not commands.Context else self.channel, data=r)
        if kwargs.get("delete_after") is not None:
            await msg.delete(delay=kwargs.get("delete_after"))
    
        return msg

def message_override(cls, *args, **kwargs):
    if cls is discord.message.Message:
        return object.__new__(Message)
    else:
        return object.__new__(cls)


module.abc.Messageable.send = send
module.message.Message.__new__ = message_override
#endregion

#region webhook override
def webhook_message_override(cls, *args, **kwargs):
    if cls is discord.webhook.WebhookMessage:
        return object.__new__(WebhookMessage)
    else:
        return object.__new__(cls)

def send_webhook(self: discord.Webhook, content=MISSING, *, wait=False, username=MISSING, avatar_url=MISSING, tts=False, files=None, embed=MISSING, embeds=MISSING, allowed_mentions=MISSING, components=MISSING):
    payload = jsonifyMessage(content, tts=tts, embed=embed, embeds=embeds, allowed_mentions=allowed_mentions, components=components)

    if username is not None:
        payload["username"] = username
    if avatar_url is not None:
        payload["avatar_url"] = str(avatar_url)
    
    return self._adapter.execute_webhook(payload=payload, wait=wait, files=files)

module.webhook.Webhook.send = send_webhook
module.webhook.WebhookMessage.__new__ = webhook_message_override
#endregion



class Overriden_Bot(commands.bot.Bot):
    """
    Represents an overriden discord bot.

    This only adds two attributes, `.slash` and `.components`

    This class is a subclass of :class:`discord.Client` and as a result
    anything that you can do with a :class:`discord.Client` you can do with
    this bot.

    This class also subclasses :class:`.GroupMixin` to provide the functionality
    to manage commands.

    Attributes
    -----------
    command_prefix
        The command prefix is what the message content must contain initially
        to have a command invoked. This prefix could either be a string to
        indicate what the prefix should be, or a callable that takes in the bot
        as its first parameter and :class:`discord.Message` as its second
        parameter and returns the prefix. This is to facilitate "dynamic"
        command prefixes. This callable can be either a regular function or
        a coroutine.

        An empty string as the prefix always matches, enabling prefix-less
        command invocation. While this may be useful in DMs it should be avoided
        in servers, as it's likely to cause performance issues and unintended
        command invocations.

        The command prefix could also be an iterable of strings indicating that
        multiple checks for the prefix should be used and the first one to
        match will be the invocation prefix. You can get this prefix via
        :attr:`.Context.prefix`. To avoid confusion empty iterables are not
        allowed.

        .. note::

            When passing multiple prefixes be careful to not pass a prefix
            that matches a longer prefix occurring later in the sequence.  For
            example, if the command prefix is ``('!', '!?')``  the ``'!?'``
            prefix will never be matched to any message as the previous one
            matches messages starting with ``!?``. This is especially important
            when passing an empty string, it should always be last as no prefix
            after it will be matched.
    case_insensitive: :class:`bool`
        Whether the commands should be case insensitive. Defaults to ``False``. This
        attribute does not carry over to groups. You must set it to every group if
        you require group commands to be case insensitive as well.
    description: :class:`str`
        The content prefixed into the default help message.
    help_command: Optional[:class:`.HelpCommand`]
        The help command implementation to use. This can be dynamically
        set at runtime. To remove the help command pass ``None``. For more
        information on implementing a help command, see :ref:`ext_commands_help_command`.
    owner_id: Optional[:class:`int`]
        The user ID that owns the bot. If this is not set and is then queried via
        :meth:`.is_owner` then it is fetched automatically using
        :meth:`~.Bot.application_info`.
    owner_ids: Optional[Collection[:class:`int`]]
        The user IDs that owns the bot. This is similar to :attr:`owner_id`.
        If this is not set and the application is team based, then it is
        fetched automatically using :meth:`~.Bot.application_info`.
        For performance reasons it is recommended to use a :class:`set`
        for the collection. You cannot set both ``owner_id`` and ``owner_ids``.

        .. versionadded:: 1.3
    strip_after_prefix: :class:`bool`
        Whether to strip whitespace characters after encountering the command
        prefix. This allows for ``!   hello`` and ``!hello`` to both work if
        the ``command_prefix`` is set to ``!``. Defaults to ``False``.

        .. versionadded:: 1.7
    """
    def __init__(self, command_prefix, help_command = None, description = None, slash_options = None, **options):
        commands.bot.Bot.__init__(self, command_prefix, help_command=help_command, description=description, **options)
        
        self.slash = Slash(self, slash_options)
        self.components = Components(self)

def client_override(cls, *args, **kwargs):
    if cls is commands.bot.Bot:
        return object.__new__(Overriden_Bot)
    else:
        return object.__new__(cls)

def override_client():
    """Overrides the default :class:`discord.ext.commands.Bot` client with a custom one, 
    which automatically initalizes the :class:`~Slash` and :class:`~Component` classes and adds them
    to the client's attributes

    Without overriding
    
    ```py
        from discord.ext import commands
        from discord_ui import UI
        
        client = commands.Bot(...)
        ui = UI(client)
    ```

    With overriding

    ```py

        from discord.ext import commands
        from discord_ui import override_client
        
        override_client()
        client = commands.Bot(...)
    ```

    ``client.slash`` and ``client.components`` are now accesable, if you want intellisense to work, change it to

    ```py

        from discord.ext import commands
        from discord_ui import override_client, OverridenClient

        override_client()
        client: OverridenClient = commands.Bot(...)
    ```

    Or you could use

    ```py

        from discord_ui import OverridenClient
        client = OverridenClient(...)
    ```

    it would still result in the same
    """
    module.ext.commands.bot.Bot.__new__ = client_override


sys.modules["discord"] = module



from .client import Components, Slash, UI
from .components import ActionRow, Button, LinkButton, SelectMenu, SelectOption, Colors
from .slash.types import OptionType, SlashPermission, SlashOption
from .slash.tools import ParseMethod
from .tools import components_to_dict
from .receive import ResponseMessage, Message, WebhookMessage, PressedButton, SelectedMenu, SlashedCommand, SlashedSubCommand, EphemeralComponent, EphemeralMessage
from .override import Overriden_Bot, override_client


__version__ = "3.0.0"
